from __future__ import annotations

import asyncio
import collections
import contextlib
import random
import threading
from abc import ABC
from asyncio import Event, QueueFull, get_event_loop
from collections.abc import Iterator
from types import GenericAlias
from typing import NoReturn

from pylav.helpers.singleton import synchronized_method_call_with_self_threading_lock
from pylav.type_hints.generics import ANY_GENERIC_TYPE


class PlayerQueue(asyncio.Queue[ANY_GENERIC_TYPE]):
    """A queue, useful for coordinating producer and consumer coroutines.

    If maxsize is less than or equal to zero, the queue size is infinite. If it
    is an integer greater than 0, then "await put()" will block when the
    queue reaches maxsize, until an item is removed by get().

    Unlike the standard library Queue, you can reliably know this Queue's size
    with qsize(), since your single-threaded asyncio application won't be
    interrupted between calling qsize() and doing an operation on the Queue.
    """

    __slots__ = ("_queue", "_maxsize", "_getters", "_putters", "_unfinished_tasks", "_finished", "_loop")

    _queue: collections.deque[ANY_GENERIC_TYPE]
    raw_b64s: list[str]

    def __init__(self, maxsize: int = 0) -> None:
        self._threading_lock = threading.Lock()
        super().__init__(maxsize=maxsize)
        self._maxsize = maxsize
        self._loop = get_event_loop()

        # Futures.
        self._getters: collections.deque[asyncio.Future] = collections.deque()
        # Futures.
        self._putters: collections.deque[asyncio.Future] = collections.deque()
        self._unfinished_tasks = 0
        self._finished = Event()
        self._finished.set()
        self._init(maxsize)

    # These do not exist in asyncio.Queue
    @property
    def raw_queue(self) -> collections.deque[ANY_GENERIC_TYPE]:
        return self._queue.copy()

    @raw_queue.setter
    def raw_queue(self, value: collections.deque[ANY_GENERIC_TYPE]):
        if not isinstance(value, collections.deque):
            raise TypeError("Queue value must be a collections.deque[Track]")
        if self._maxsize and len(value) > self._maxsize:
            raise ValueError(f"Queue value cannot be longer than maxsize: {self._maxsize}")
        self._queue = value

    @raw_queue.deleter
    def raw_queue(self) -> None:
        self.clear()

    @synchronized_method_call_with_self_threading_lock()
    def popindex(self, index: int) -> ANY_GENERIC_TYPE:
        value = self._queue[index]
        del self._queue[index]
        return value

    @synchronized_method_call_with_self_threading_lock()
    async def remove(self, value: ANY_GENERIC_TYPE, duplicates: bool = False) -> tuple[list[ANY_GENERIC_TYPE], int]:
        """Removes the first occurrence of a value from the queue.

        If duplicates is True, all occurrences of the value are removed.
        Returns the removed entries and number of occurrences removed.
        """
        count = 0
        removed = []
        try:
            i = self._queue.index(value)
            removed.append(self.popindex(i))
            count += 1
            if duplicates:
                with contextlib.suppress(ValueError):
                    while value in self:
                        i = self._queue.index(value)
                        removed.append(self.popindex(i))
                        count += 1
            return removed, count
        except ValueError as e:
            raise IndexError("Value not in queue") from e

    @synchronized_method_call_with_self_threading_lock()
    def clear(self) -> None:
        """Remove all items from the queue"""
        self._queue.clear()
        for i in self._getters:
            i.cancel()
        self._getters.clear()
        for i in self._putters:
            i.cancel()
        self._putters.clear()

    @synchronized_method_call_with_self_threading_lock()
    async def shuffle(self) -> None:
        """Shuffle the queue"""
        if self.empty():
            return
        await asyncio.to_thread(random.shuffle, self._queue)

    @synchronized_method_call_with_self_threading_lock()
    async def get_oldest(self) -> ANY_GENERIC_TYPE:
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.
        """
        while self.empty():
            getter = self._loop.create_future()
            self._getters.append(getter)
            try:
                await getter
            except BaseException:
                getter.cancel()
                with contextlib.suppress(ValueError):
                    self._getters.remove(getter)
                if not self.empty() and not getter.cancelled():
                    self._wakeup_next(self._getters)
                raise
        return self.get_nowait(index=-1)

    def index(self, value: ANY_GENERIC_TYPE) -> int:
        """Return first index of value"""
        return self._queue.index(value)

    def __contains__(self, obj: ANY_GENERIC_TYPE) -> bool:
        return obj in self._queue

    def __iter__(self) -> Iterator[ANY_GENERIC_TYPE]:
        return iter(self.raw_queue)

    def __len__(self) -> int:
        return len(self._queue)

    def __index__(self) -> int:
        return len(self._queue)

    def __getitem__(self, key: int | slice) -> ANY_GENERIC_TYPE | list[ANY_GENERIC_TYPE]:
        return self.raw_queue[key]

    def __setitem__(self, key: int, value: ANY_GENERIC_TYPE | list[ANY_GENERIC_TYPE]) -> NoReturn:
        raise NotImplementedError("Use .put() to add entries to the queue")

    def __delitem__(self, key: int) -> NoReturn:
        raise NotImplementedError("Use .get() to remove entries from the queue")

    def __length_hint__(self) -> int:
        return len(self._queue)

    # These three are overridable in subclasses.

    def _init(self, maxsize: int) -> None:
        self._queue = collections.deque(maxlen=maxsize or None)
        self.raw_b64s = []

    @synchronized_method_call_with_self_threading_lock()
    def _get(self, index: int = None) -> ANY_GENERIC_TYPE:
        r = self.popindex(index) if index is not None else self._queue.popleft()
        if r.encoded:
            self.raw_b64s.remove(r.encoded)
        return r

    @synchronized_method_call_with_self_threading_lock()
    def _put(self, items: list[ANY_GENERIC_TYPE], index: int = None) -> None:
        if index is not None:
            for i in items:
                if index < 0:
                    self._queue.append(i)
                    if i.encoded:
                        self.raw_b64s.append(i.encoded)
                else:
                    self._queue.insert(index, i)
                    if i.encoded:
                        self.raw_b64s.append(i.encoded)
                    index += 1
        else:
            self._queue.extend(items)
            self.raw_b64s.extend([i.encoded for i in items if i.encoded])

    # End of the overridable methods.

    @staticmethod
    def _wakeup_next(waiters: collections.deque[asyncio.Future]) -> None:
        # Wake up the next waiter (if any) that isn't cancelled.
        while waiters:
            waiter = waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                break

    def __repr__(self) -> str:
        return f"<{type(self).__name__} at {id(self):#x} {self._format()}>"

    def __str__(self) -> str:
        return f"<{type(self).__name__} {self._format()}>"

    __class_getitem__ = classmethod(GenericAlias)

    def _format(self) -> str:
        result = f"maxsize={self._maxsize!r}"
        if getattr(self, "_queue", None):
            result += f" _queue={list(self._queue)!r}"
        if self._getters:
            result += f" _getters[{len(self._getters)}]"
        if self._putters:
            result += f" _putters[{len(self._putters)}]"
        if self._unfinished_tasks:
            result += f" tasks={self._unfinished_tasks}"
        return result

    def qsize(self) -> int:
        """Number of items in the queue"""
        return len(self._queue)

    size = qsize

    @property
    def maxsize(self) -> int:
        """Number of items allowed in the queue"""
        return self._maxsize

    def empty(self) -> bool:
        """Return True if the queue is empty, False otherwise"""
        return not len(self._queue)

    def full(self) -> bool:
        """Return True if there are maxsize items in the queue.

        Note: if the Queue was initialized with maxsize=0 (the default),
        then full() is never True.
        """
        return False if self._maxsize <= 0 else self.qsize() >= self._maxsize

    async def put(self, items: list[ANY_GENERIC_TYPE], index: int = None, discard: bool = False) -> None:
        """Put an item into the queue.

        Put an item into the queue. If the queue is full, wait until a free
        slot is available before adding item.
        """
        if discard:
            while self.full():
                await self.get_oldest()
        while self.full():
            putter = self._loop.create_future()
            self._putters.append(putter)
            try:
                await putter
            except BaseException:
                putter.cancel()  # Just in case putter is not done yet.
                with contextlib.suppress(ValueError):
                    # Clean self._putters from canceled putters.
                    self._putters.remove(putter)
                if not self.full() and not putter.cancelled():
                    # We were woken up by get_nowait(), but can't take
                    # the call.  Wake up the next in line.
                    self._wakeup_next(self._putters)
                raise
        return self.put_nowait(items, index)

    def put_nowait(self, items: list[ANY_GENERIC_TYPE], index: int = None) -> None:
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        if self.full():
            raise QueueFull
        self._put(items, index=index)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    async def get(self, index: int = None) -> ANY_GENERIC_TYPE:
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.
        """
        while self.empty():
            getter = self._loop.create_future()
            self._getters.append(getter)
            try:
                await getter
            except BaseException:
                getter.cancel()
                with contextlib.suppress(ValueError):
                    self._getters.remove(getter)
                if not self.empty() and not getter.cancelled():
                    self._wakeup_next(self._getters)
                raise
        return self.get_nowait(index=index)

    def get_nowait(self, index: int = None) -> ANY_GENERIC_TYPE:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        if self.empty():
            return None
        item = self._get(index=index)
        self._wakeup_next(self._putters)
        return item

    def task_done(self) -> None:
        """Indicate that a formerly enqueued task is complete.

        Used by queue consumers. For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items have
        been processed (meaning that a task_done() call was received for every
        item that had been put() into the queue).

        Raises ValueError if called more times than there were items placed in
        the queue.
        """
        if self._unfinished_tasks <= 0:
            raise ValueError("task_done() called too many times")
        self._unfinished_tasks -= 1
        if self._unfinished_tasks == 0:
            self._finished.set()

    async def join(self) -> None:
        """Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls task_done() to
        indicate that the item was retrieved and all work on it is complete.
        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        if self._unfinished_tasks > 0:
            await self._finished.wait()


class TrackHistoryQueue(PlayerQueue[ANY_GENERIC_TYPE], ABC):
    __slots__ = ("_queue", "_maxsize", "_getters", "_putters", "_unfinished_tasks", "_finished", "_loop")

    def __int__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize=maxsize)

    def _init(self, maxsize: int) -> None:
        super()._init(maxsize)

    def _put(self, items: list[ANY_GENERIC_TYPE], index: int = None) -> None:
        if len(items) + self.qsize() > self.maxsize:
            diff = len(items) + self.qsize() - self.maxsize
            for _ in range(diff):
                self._queue.pop()
                self.raw_b64s.pop()
        if index is not None:
            for i in items:
                i.timestamp = 0
                if index < 0:
                    self._queue.append(i)
                    self.raw_b64s.append(i.encoded)
                else:
                    self._queue.insert(index, i)
                    self.raw_b64s.insert(index, i.encoded)
                    index += 1
        else:
            self._queue.extendleft(items)
            for i, t in enumerate(items):
                t.timestamp = 0
                self.raw_b64s.insert(i, t.encoded)

    def _get(self, index: int = None) -> ANY_GENERIC_TYPE:
        r = self.popindex(index) if index is not None else self._queue.popleft()
        self.raw_b64s.pop(index if index is not None else -1)
        return r

    def put_nowait(self, items: list[ANY_GENERIC_TYPE], index: int = None) -> None:
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._put(items, index=index)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    def get_nowait(self, index: int = None) -> ANY_GENERIC_TYPE | None:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        if self.empty():
            return None
        item = self._get(index=index)
        self._wakeup_next(self._putters)
        return item



from __future__ import annotations

import logging
import os

from red_commons.logging import RedTraceLogger  # type: ignore
from red_commons.logging import getLogger as redgetLogger
from red_commons.logging import maybe_update_logger_class

__all__ = ("getLogger",)

maybe_update_logger_class()

LOGGER_PREFIX = os.getenv("PYLAV__LOGGER_PREFIX", "")
__deepdiff = logging.getLogger("deepdiff.diff")
__deepdiff.setLevel(logging.CRITICAL)
__deepdiff.propagate = False
__deepdiff.disabled = True
__watchfiles = logging.getLogger("watchfiles")
__watchfiles.setLevel(logging.CRITICAL)
__watchfiles.propagate = False
__watchfiles.disabled = True
logging.getLogger("apscheduler").setLevel(logging.ERROR)
logging.getLogger("aiohttp_client_cache").setLevel(logging.ERROR)


# noinspection PyPep8Naming
def getLogger(name: str) -> RedTraceLogger:  # noqa: N802
    return redgetLogger(f"{LOGGER_PREFIX}{name}")



from __future__ import annotations

import dataclasses
from typing import Literal  # noqa

from pylav.nodes.api.responses.exceptions import LavalinkException as TrackExceptionClass
from pylav.nodes.api.responses.player import State
from pylav.nodes.api.responses.track import Track


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class CPU:
    cores: int
    systemLoad: float
    lavalinkLoad: float


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Memory:
    free: int
    allocated: int
    reservable: int
    used: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Frame:
    sent: int
    nulled: int
    deficit: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Message:
    op: str


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Ready(Message):
    sessionId: str
    resumed: bool


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Stats(Message):
    players: int
    playingPlayers: int
    uptime: int
    memory: Memory
    cpu: CPU
    frameStats: Frame | None = None
    uptime_seconds: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "uptime_seconds", self.uptime / 1000)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlayerUpdate(Message):
    guildId: str
    state: State


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackStart(Message):
    op: Literal["event"]
    type: Literal["TrackStartEvent"]
    guildId: str
    track: Track


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackStuck(Message):
    op: Literal["event"]
    type: Literal["TrackStuckEvent"]
    guildId: str
    track: Track
    thresholdMs: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackEnd(Message):
    op: Literal["event"]
    type: Literal["TrackEndEvent"]
    guildId: str
    track: Track
    reason: Literal["FINISHED", "LOAD_FAILED", "STOPPED", "REPLACED", "CLEANUP"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackException(Message):
    op: Literal["event"]
    type: Literal["TrackExceptionEvent"]
    guildId: str
    track: Track
    exception: TrackExceptionClass


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Closed(Message):
    op: Literal["event"]
    type: Literal["WebSocketClosedEvent"]
    guildId: str
    code: int
    reason: str
    byRemote: bool



from __future__ import annotations

import inspect
import pathlib
import sys
from collections.abc import Callable
from typing import TypeVar, get_type_hints

from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE

T = TypeVar("T")


def _get_path(path: T | pathlib.Path) -> str | T | None:
    from pylav.extension.bundled_node.utils import get_true_path

    return get_true_path(path, fallback=path)


def update_event_loop_policy() -> None:
    if sys.implementation.name == "cpython":
        # Let's not force this dependency, uvloop is much faster on cpython
        try:
            import uvloop  # type: ignore
        except ImportError:
            pass
        else:
            import asyncio

            if not isinstance(asyncio.get_event_loop_policy(), uvloop.EventLoopPolicy):
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def check_annotated(
    func: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
    hints = get_type_hints(func, include_extras=True)
    spec = inspect.getfullargspec(func)

    def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> ANY_GENERIC_TYPE:
        for idx, arg_name in enumerate(spec[0]):
            hint = hints.get(arg_name)
            validators = getattr(hint, "__metadata__", None)
            if not validators:
                continue
            for validator in validators:
                validator.validate_value(args[idx])

        return func(*args, **kwargs)

    return wrapper


def add_property(inst: object, name: str, method: Callable) -> None:
    cls = type(inst)
    if not hasattr(cls, "__per_instance"):
        cls = type(cls.__name__, (cls,), {})
        cls.__per_instance = True
        inst.__class__ = cls
    setattr(cls, name, property(method))



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.constants.versions import VERSION_0_0_0, VERSION_0_10_5_0
from pylav.extension.bundled_node import LAVALINK_DOWNLOAD_DIR
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_0_10_5_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_0_10_5_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_0_10_5_0)
    config = client.node_db_manager.bundled_node_config()
    yaml_data = await config.fetch_yaml()
    plugins = yaml_data["lavalink"]["plugins"]
    keep = [
        plugin
        for plugin in plugins
        if not plugin["dependency"].startswith("com.github.Topis-Lavalink-Plugins:Topis-Source-Managers-Plugin:")
    ]

    keep.extend(
        [
            plugin
            for plugin in NODE_DEFAULT_SETTINGS["lavalink"]["plugins"]
            if plugin["dependency"].startswith("com.github.TopiSenpai.LavaSrc:lavasrc-plugin")
        ]
    )
    yaml_data["lavalink"]["plugins"] = keep
    if "topissourcemanagers" in yaml_data["plugins"]:
        yaml_data["plugins"]["lavasrc"] = yaml_data["plugins"]["topissourcemanagers"]
        del yaml_data["plugins"]["topissourcemanagers"]
    if "dzisrc:%ISRC%" not in yaml_data["plugins"]["lavasrc"]["providers"]:
        yaml_data["plugins"]["lavasrc"]["providers"].append("dzisrc:%ISRC%")
    yaml_data["plugins"]["lavasrc"]["sources"]["deezer"] = NODE_DEFAULT_SETTINGS["plugins"]["lavasrc"]["sources"][
        "deezer"
    ]
    yaml_data["plugins"]["lavasrc"]["deezer"] = NODE_DEFAULT_SETTINGS["plugins"]["lavasrc"]["deezer"]

    await config.update_yaml(yaml_data)

    folder = LAVALINK_DOWNLOAD_DIR / "plugins"
    if await folder.exists():
        plugin_files = [
            x
            async for x in folder.iterdir()
            if x.name.startswith("Topis-Source-Managers-Plugin-") and x.suffix == ".jar" and x.is_file()
        ]
        for file in plugin_files:
            try:
                await file.unlink()
            except Exception as exc:
                LOGGER.error("Failed to remove plugin: %s", file.name, exc_info=exc)

    await client.lib_db_manager.update_bot_dv_version(VERSION_0_10_5_0)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from pylav.events import base, node, player, plugins, queue, track
from pylav.events.plugins import sponsorblock
from pylav.events.track import track_start
from pylav.events.utils import get_event_name, get_simple_event_name

if TYPE_CHECKING:
    from pylav.core.client import Client


class DispatchManager:
    """
    The Dispatcher is responsible for dispatching events to the appropriate
    handlers.

    The method names are the event names.

    You can listen to events by adding the following to your client:


    Examples
    --------
    >>> from discord.ext import commands

    >>> @commands.Cog.listener()
    >>> async def on_pylav_queue_end_event(self, event: queue.QueueEndEvent):
    >>>    print(f"Queue ended: {event.player}")

    >>> @commands.Cog.listener()
    >>> async def on_pylav_track_stuck_event(self, event: track.TrackStuckEvent):
    >>>    print(f"Track got stuck: {event.track}")

    """

    __slots__ = ("_client", "dispatcher", "mapping")

    def __init__(self, client: Client) -> None:
        self._client = client
        self.dispatcher = client.bot.dispatch

        self.mapping: dict[type[base.PyLavEvent], str] = {}
        self._update_mapper(player)
        self._update_mapper(node)
        self._update_mapper(queue)
        self._update_mapper(track)
        self._update_mapper(plugins)
        self._update_mapper(track_start)
        self._update_mapper(sponsorblock)

    def _update_mapper(self, module: node | player | queue | track) -> None:  # type: ignore
        self.mapping.update(
            {
                c: get_event_name(c)
                for _, c in inspect.getmembers(module, inspect.isclass)
                if issubclass(c, base.PyLavEvent)
            }
        )

    async def dispatch(self, event: base.PyLavEvent) -> None:
        event_name = self.mapping[type(event)]
        self.dispatcher(event_name, event)

    def get_event_names(self) -> set[str]:
        return set(self.mapping.values())

    def simple_event_names(self) -> set[str]:
        return {get_simple_event_name(k) for k in self.mapping.keys()}



from __future__ import annotations

import heapq
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from discord.app_commands import Choice, Transformer
from discord.ext import commands
from rapidfuzz import fuzz

from pylav.exceptions.database import EntryNotFoundException
from pylav.helpers.format.strings import shorten_string
from pylav.nodes.node import Node
from pylav.type_hints.bot import DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    NodeConverter = TypeVar("NodeConverter", bound=list[Node])
else:

    class NodeConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Node]:
            """Converts a node name or ID to a list of matching objects"""
            try:
                nodes = ctx.pylav.node_manager.nodes
            except EntryNotFoundException as e:
                raise commands.BadArgument(
                    _("Node with name or identifier `{user_input_variable_do_not_translate}` not found.").format(
                        user_input_variable_do_not_translate=arg
                    )
                ) from e
            if r := list(filter(lambda n: arg.lower() in n.name.lower() or arg == f"{n.identifier}", nodes)):
                return r
            raise commands.BadArgument(
                _("Node with name or identifier `{user_input_variable_do_not_translate}` not found").format(
                    user_input_variable_do_not_translate=arg
                )
            )

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Node]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            nodes = interaction.client.pylav.node_manager.nodes
            if not current:
                return [Choice(name=shorten_string(e.name, max_length=100), value=f"{e.identifier}") for e in nodes][
                    :25
                ]

            def _filter(c):
                return fuzz.partial_ratio(c.name, current)

            extracted = heapq.nlargest(25, nodes, key=_filter)

            return [Choice(name=shorten_string(e.name, max_length=100), value=f"{e.identifier}") for e in extracted]



from __future__ import annotations

import asyncio
import contextlib
import datetime
import pathlib
import platform
import re
import shlex
import shutil
import tempfile
from typing import TYPE_CHECKING, Any

import aiohttp
import aiopath
import dateutil
import dateutil.parser
import psutil
import rich.progress
import yaml
from packaging.version import Version

from pylav.compat import json
from pylav.constants.config import JAVA_EXECUTABLE
from pylav.constants.misc import EPOCH_DT_TZ_AWARE
from pylav.constants.node import JAR_SERVER_RELEASES
from pylav.constants.regex import (
    JAVA_VERSION_LINE_223,
    JAVA_VERSION_LINE_PRE223,
    LAVALINK__READY_LINE,
    LAVALINK_BRANCH_LINE,
    LAVALINK_BUILD_LINE,
    LAVALINK_BUILD_TIME_LINE,
    LAVALINK_COMMIT_LINE,
    LAVALINK_FAILED_TO_START,
    LAVALINK_JAVA_LINE,
    LAVALINK_LAVAPLAYER_LINE,
    LAVALINK_VERSION_LINE,
    SEMANTIC_VERSION_LAZY,
)
from pylav.constants.versions import VERSION_4_0_0
from pylav.exceptions.node import (
    EarlyExitException,
    IncorrectProcessFoundException,
    InvalidArchitectureException,
    LavalinkDownloadFailedException,
    ManagedLavalinkNodeException,
    ManagedLavalinkStartFailureException,
    ManagedLinkStartAbortedUseExternal,
    NodeUnhealthyException,
    NoProcessFoundException,
    PortAlreadyInUseException,
    TooManyProcessFoundException,
    UnexpectedJavaResponseException,
    UnsupportedJavaException,
    WebsocketNotConnectedException,
)
from pylav.extension.bundled_node import LAVALINK_APP_YML, LAVALINK_DOWNLOAD_DIR, LAVALINK_JAR_FILE, USING_FORCED
from pylav.extension.bundled_node.utils import change_dict_naming_convention, get_jar_ram_actual, get_true_path
from pylav.helpers.misc import ExponentialBackoffWithReset
from pylav.logging import getLogger
from pylav.nodes.node import Node
from pylav.storage.migrations.high_level.always.update_plugins import update_plugins
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    from pylav.core.client import Client


LOGGER = getLogger("PyLav.ManagedNode")


class LocalNodeManager:
    """A manager for a local Lavalink node."""

    __slots__ = (
        "ready",
        "_ci_info",
        "_client",
        "_proc",
        "_node_pid",
        "_shutdown",
        "start_monitor_task",
        "timeout",
        "_args",
        "_session",
        "_node_id",
        "_node",
        "_current_config",
        "abort_for_unmanaged",
        "_wait_for",
        "_java_path",
        "_java_exc",
        "_java_available",
        "_java_version",
        "_java_version",
        "_up_to_date",
        "_blacklisted_archs",
        "_lavaplayer",
        "_lavalink_build",
        "_jvm",
        "_lavalink_branch",
        "_buildtime",
        "_commit",
        "_version",
        "__buffer_task",
        "_disabled",
    )

    def __init__(self, client: Client, timeout: int | None = None) -> None:
        self._java_available: bool | None = None
        self._java_version: tuple[int, int] | None = None
        self._up_to_date: bool | None = None
        self._blacklisted_archs: list[str] = []

        self._lavaplayer: str | None = None
        self._lavalink_build: int | None = None
        self._jvm: str | None = None
        self._lavalink_branch: str | None = None
        self._buildtime: str | None = None
        self._commit: str | None = None
        self._version: str | None = None
        self._java_exc: str = JAVA_EXECUTABLE

        self.ready: asyncio.Event = asyncio.Event()
        self._ci_info: dict = {"number": 0, "branchName": "", "finishDate": "", "href": "", "jar_url": ""}
        self._client = client
        self._proc: asyncio.subprocess.Process | None = None  # pylint:disable=no-member
        self._node_pid: int | None = None
        self._shutdown: bool = False
        self.start_monitor_task = None
        self.timeout = timeout
        self._args = []
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=json.dumps)
        self._node_id: int = self._client.bot.user.id
        self._node: Node | None = None
        self._current_config = {}
        self.abort_for_unmanaged: asyncio.Event = asyncio.Event()
        self._args = []
        self._wait_for = asyncio.Event()
        self._java_path = None
        self.__buffer_task = None
        self._disabled = True

    @property
    def disabled(self) -> bool:
        return self._disabled

    @property
    def node(self) -> Node | None:
        return self._node

    @property
    def path(self) -> str | None:
        return self._java_exc

    @property
    def jvm(self) -> str | None:
        return self._jvm

    @property
    def lavaplayer(self) -> str | None:
        return self._lavaplayer

    @property
    def ll_build(self) -> int | None:
        return self._lavalink_build

    @property
    def ll_branch(self) -> str | None:
        return self._lavalink_branch

    @property
    def build_time(self) -> str | None:
        return self._buildtime

    @staticmethod
    def _get_release_publish_dt_or_epoch(release: dict) -> datetime.datetime:
        return (
            dateutil.parser.parse(release["published_at"])
            if Version(release["tag_name"]) < VERSION_4_0_0
            else EPOCH_DT_TZ_AWARE
        )

    async def get_ci_latest_info(self) -> dict[str, int | str | None]:
        async with self._client.cached_session.get(
            f"{JAR_SERVER_RELEASES}", headers={"Accept": "application/json"}
        ) as response:
            if response.status != 200:
                LOGGER.warning("Failed to get latest CI info from GitHub: %s", response.status)
                self._ci_info["number"] = -1
                return self._ci_info
            data = await response.json(loads=json.loads)
            release = max(data, key=self._get_release_publish_dt_or_epoch)
            assets = release.get("assets", [])
            url = None
            for asset in iter(assets):
                if asset["name"] != "Lavalink.jar":
                    continue
                url = asset.get("browser_download_url")
            if url is None:
                LOGGER.warning("Failed to get a supported released from from GitHub")
                self._ci_info["number"] = -1
                return self._ci_info
            date = release.get("published_at")
            branch = release.get("target_commitish")
            return {"number": int(release.get("id")), "branchName": branch, "finishDate": date, "jar_url": url}

    async def _start(self, java_path: str) -> None:
        arch_name = platform.machine()
        self._java_exc = java_path
        if arch_name in self._blacklisted_archs:
            raise InvalidArchitectureException(
                _("You are attempting to run the managed Lavalink node on an unsupported machine architecture.")
            )
        await self.process_settings()
        possible_lavalink_processes = await self.get_lavalink_process(lazy_match=True)
        if possible_lavalink_processes:
            await self.process_existing_lavalink_processes(possible_lavalink_processes)

        await self.maybe_download_jar()
        args, msg = await self._get_jar_args()
        if msg is not None:
            LOGGER.warning(msg)
        command_string = shlex.join(args)
        LOGGER.info("Managed Lavalink node startup command: %s", command_string)
        if "-Xmx" not in command_string and msg is None:
            LOGGER.warning("Managed Lavalink node maximum allowed RAM not set or higher than available RAM")
        try:
            self._proc = await asyncio.subprocess.create_subprocess_exec(  # pylint:disable=no-member
                *args,
                cwd=str(LAVALINK_DOWNLOAD_DIR),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            self._node_pid = self._proc.pid
            LOGGER.info("Managed Lavalink node started. PID: %s", self._node_pid)
            try:
                await asyncio.wait_for(self._wait_for_launcher(), timeout=self.timeout)
            except asyncio.TimeoutError:
                LOGGER.warning("Timeout occurred whilst waiting for managed Lavalink node to be ready")
                raise
        except asyncio.TimeoutError:
            await self._partial_shutdown()
        except Exception:  # noqa
            await self._partial_shutdown()
            raise

    async def process_existing_lavalink_processes(self, possible_lavalink_processes: list[dict[str, Any]]) -> None:
        LOGGER.info(
            "Found %s processes that match potential unmanaged Lavalink nodes", len(possible_lavalink_processes)
        )
        valid_working_dirs = [
            cwd  # noqa
            for d in iter(possible_lavalink_processes)
            if d.get("name") in ["java", "java.exe"] and (cwd := d.get("cwd"))
        ]
        LOGGER.debug("Found %s java processed with a cwd set", len(valid_working_dirs))
        for cwd in iter(valid_working_dirs):
            config = aiopath.AsyncPath(cwd) / "application.yml"
            if await config.exists() and await config.is_file():
                LOGGER.debug(
                    "The following settings file exists for an unmanaged Lavalink node %s",
                    config,
                )
                try:
                    async with config.open(mode="r") as config_data:
                        data = yaml.safe_load(await config_data.read())
                        data["server"]["address"]  # noqa
                        data["server"]["port"]  # noqa
                        data["lavalink"]["server"]["password"]  # noqa
                        self._node_pid = 0
                        self._current_config = data
                        raise ManagedLinkStartAbortedUseExternal
                except ManagedLinkStartAbortedUseExternal:
                    raise
                except Exception:  # noqa
                    LOGGER.exception("Failed to read contents of %s", config)
                    continue

    async def process_settings(self) -> None:
        data = await self._client.node_db_manager.bundled_node_config().fetch_yaml()
        data = change_dict_naming_convention(data)
        # The reason this is here is to completely remove these keys from the application.yml
        # if they are set to empty values
        await self.maybe_remove_youtube_config(data)
        await self.maybe_remove_ratelimit_config(data)
        await self.update_dns_config(data)
        await self.maybe_remove_proxy_config(data)
        await self.maybe_remove_apple_music_config(data)
        await self.maybe_remove_yandex_config(data)
        data["plugins"]["lavasrc"]["providers"] = list(dict.fromkeys(data["plugins"]["lavasrc"]["providers"]))

        await self.maybe_update_apple_music_country_code(data)
        await self.maybe_update_spotify_country_code(data)
        await self.maybe_update_tts_country_code(data)

        self._current_config = data
        async with LAVALINK_APP_YML.open("w") as f:
            await f.write(yaml.safe_dump(data))

    @staticmethod
    async def maybe_update_tts_country_code(data: JSON_DICT_TYPE) -> None:
        if len(data["plugins"]["dunctebot"]["ttsLanguage"]) != 5:
            data["plugins"]["dunctebot"]["ttsLanguage"] = "en-US"
            LOGGER.warning("Invalid TTS language code provided for dunctebot plugin, defaulting to en-US")

    @staticmethod
    async def maybe_update_spotify_country_code(data: JSON_DICT_TYPE) -> None:
        if len(data["plugins"]["lavasrc"]["spotify"]["countryCode"]) != 2:
            data["plugins"]["lavasrc"]["spotify"]["countryCode"] = "US"
            LOGGER.warning(
                "Spotify country code is not set to a valid ISO 3166-1 alpha-2 code and has been defaulted to US"
            )

    @staticmethod
    async def maybe_update_apple_music_country_code(data: JSON_DICT_TYPE) -> None:
        if len(data["plugins"]["lavasrc"]["applemusic"]["countryCode"]) != 2:
            data["plugins"]["lavasrc"]["applemusic"]["countryCode"] = "US"
            LOGGER.warning(
                "Apple Music country code is not set to a valid ISO 3166-1 alpha-2 code and has been defaulted to US"
            )

    @staticmethod
    async def maybe_remove_yandex_config(data: JSON_DICT_TYPE) -> None:
        if (
            "accessToken" not in data["plugins"]["lavasrc"]["yandexmusic"]
            or not data["plugins"]["lavasrc"]["yandexmusic"]["accessToken"]
        ):
            del data["plugins"]["lavasrc"]["yandexmusic"]

    @staticmethod
    async def maybe_remove_apple_music_config(data: JSON_DICT_TYPE) -> None:
        if (
            "mediaAPIToken" not in data["plugins"]["lavasrc"]["applemusic"]
            or not data["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"]
        ):
            del data["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"]

    @staticmethod
    async def maybe_remove_proxy_config(data: JSON_DICT_TYPE) -> None:
        if not data["lavalink"]["server"]["httpConfig"].get("proxyHost"):
            del data["lavalink"]["server"]["httpConfig"]

    async def update_dns_config(self, data: JSON_DICT_TYPE) -> None:
        if data["sentry"]["dsn"]:
            data["sentry"]["tags"]["ID"] = self._client.bot.user.id
            data["sentry"]["tags"]["pylav_version"] = self._client.lib_version

    @staticmethod
    async def maybe_remove_ratelimit_config(data: JSON_DICT_TYPE) -> None:
        if not (
            data["lavalink"]["server"]["ratelimit"].get("ipBlocks")
            and data["lavalink"]["server"]["ratelimit"].get("strategy")
        ):
            del data["lavalink"]["server"]["ratelimit"]

    @staticmethod
    async def maybe_remove_youtube_config(data: JSON_DICT_TYPE) -> None:
        if not all(
            (
                data["lavalink"]["server"]["youtubeConfig"].get("email"),
                data["lavalink"]["server"]["youtubeConfig"].get("password"),
            )
        ):
            del data["lavalink"]["server"]["youtubeConfig"]

    async def _get_jar_args(self) -> tuple[list[str], str | None]:
        java_available, java_version = await self._has_java()
        if not java_available:
            raise OSError(f"Pylav - Java executable not found, tried to use: '{self._java_exc}'")

        java_xms_default, java_xmx_default, __, java_max_allowed_ram = get_jar_ram_actual(self._java_exc)
        java_xms, java_xmx = (
            java_xms_default,
            (await self._client.node_db_manager.bundled_node_config().fetch_extras()).get("max_ram", java_xmx_default),
        )

        match = re.match(r"^(\d+)([MG])$", java_xmx, flags=re.IGNORECASE)
        command_args = [self._java_exc, f"-Xms{java_xms}"]
        if (11, 0) <= java_version < (12, 0):
            command_args.append("-Djdk.tls.client.protocols=TLSv1.2")
        meta = 0, None
        invalid = None
        if match and int(match[1]) * 1024 ** (2 if match[2].lower() == "m" else 3) <= java_max_allowed_ram:
            command_args.append(f"-Xmx{java_xmx}")
        elif meta[0] is not None:
            invalid = "Managed Lavalink node RAM allocation ignored due to system limitations, please fix this"

        command_args.extend(["-jar", str(LAVALINK_JAR_FILE)])
        self._args = command_args
        return command_args, invalid

    async def _has_java(self) -> tuple[bool, tuple[int, int] | None]:
        if self._java_available:
            # Return cached value if we've checked this before
            return self._java_available, self._java_version
        java_exec = get_true_path(self._java_exc)
        java_available = java_exec is not None
        if not java_available:
            self._java_available = False
            self._java_version = None
        else:
            self._java_version = await self._get_java_version()
            self._java_available = self._java_version >= (11, 0)  # https://github.com/freyacodes/Lavalink#requirements
            self._java_exc = java_exec
        return self._java_available, self._java_version

    async def _get_java_version(self) -> tuple[int, int]:
        """This assumes we've already checked that java exists"""
        _proc: asyncio.subprocess.Process = await asyncio.create_subprocess_exec(  # pylint:disable=no-member
            self._java_exc,
            "-version",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        # java -version outputs to stderr
        __, err = await _proc.communicate()

        version_info: str = err.decode("utf-8")
        lines = version_info.splitlines()

        for line in iter(lines):
            match = JAVA_VERSION_LINE_PRE223.search(line)
            if match is None:
                match = JAVA_VERSION_LINE_223.search(line)
            if match is None:
                match = SEMANTIC_VERSION_LAZY.search(line)
            if match is None:
                continue
            major = int(match["major"])
            minor = 0
            if minor_str := match["minor"]:
                minor = int(minor_str)

            return major, minor

        raise UnexpectedJavaResponseException(
            _(
                "The output of `{command_name_variable_do_not_translate}` was unexpected\n{command_output_variable_do_not_translate}"
            ).format(
                command_name_variable_do_not_translate=f"{self._java_exc} -version",
                command_output_variable_do_not_translate=version_info,
            )
        )

    async def __consume_buffer(self) -> None:
        async for __ in self._proc.stdout:
            pass

    async def _wait_for_launcher(self) -> None:
        LOGGER.info("Waiting for Managed Lavalink node to be ready")
        async for line in self._proc.stdout:
            if LAVALINK__READY_LINE.search(line):
                self.ready.set()
                LOGGER.info("Managed Lavalink node is ready to receive requests")
                self.__buffer_task = asyncio.create_task(self.__consume_buffer())
                break
            if LAVALINK_FAILED_TO_START.search(line):
                if f"Port {self._current_config['server']['port']} was already in use".encode() in line:
                    raise PortAlreadyInUseException(
                        _(
                            "The port {port_variable_do_not_translate} is already in use. Managed Lavalink startup has been aborted."
                        ).format(port_variable_do_not_translate=self._current_config["server"]["port"])
                    )
                raise ManagedLavalinkStartFailureException(
                    _("Lavalink failed to start: {error_variable_do_not_translate}").format(
                        error_variable_do_not_translate=line.decode("utf-8")
                    )
                )
            if self._proc.returncode is not None:
                # Avoid Console spam only print once every 2 seconds
                raise EarlyExitException("Managed Lavalink node server exited early")

    async def shutdown(self) -> None:
        self._disabled = True
        if self.start_monitor_task is not None:
            self.start_monitor_task.cancel()
        if self.node:
            await self._client.node_manager.remove_node(self.node)
            self._node = None
        await self._partial_shutdown()
        await self._session.close()

    async def _partial_shutdown(self) -> None:
        self.ready.clear()
        self._wait_for.clear()
        self.abort_for_unmanaged.clear()
        # In certain situations to await self._proc.wait() is invalid so waiting on it waits forever.
        if self._shutdown is True:
            # For convenience, calling this method more than once or calling it before starting it
            # does nothing.
            return
        if self.__buffer_task is not None:
            self.__buffer_task.cancel()
            self.__buffer_task = None
        await self.maybe_kill_existing_process()
        await self.maybe_kill_alive_process()
        self._proc = None
        self._shutdown = True
        self._node_pid = None
        if self._node is not None:
            await self._client.remove_node(self._node_id)
            self._node = None

    async def maybe_kill_alive_process(self) -> None:
        if self._proc is not None and self._proc.returncode is None:
            self._proc.terminate()
            self._proc.kill()
            await self._proc.wait()

    async def maybe_kill_existing_process(self) -> None:
        if self._node_pid:
            with contextlib.suppress(psutil.Error):
                p = psutil.Process(self._node_pid)
                p.terminate()
                p.kill()

    async def should_auto_update(self) -> bool:
        # noinspection PyProtectedMember
        return (
            False
            if USING_FORCED
            else await self._client._lib_config_manager.get_config().fetch_auto_update_managed_nodes()
        )

    async def _download_jar(self, forced: bool = False) -> None:
        if not await self.should_auto_update() and not forced:
            return
        LOGGER.info("Downloading Lavalink.jar")
        jar_url = (
            self._ci_info["jar_url"] or "https://github.com/freyacodes/Lavalink/releases/download/4.0.0/Lavalink.jar"
        )

        async with self._session.get(jar_url, timeout=3600) as response:
            if response.status == 404:
                raise LavalinkDownloadFailedException(response=response, should_retry=False)
            elif 400 <= response.status < 600:
                raise LavalinkDownloadFailedException(response=response, should_retry=True)
            fd, path = tempfile.mkstemp()
            file = open(fd, "wb")
            nbytes = 0
            with rich.progress.Progress(
                rich.progress.SpinnerColumn(),
                rich.progress.TextColumn("[progress.description]{task.description}"),
                rich.progress.BarColumn(),
                rich.progress.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                rich.progress.TimeRemainingColumn(),
                rich.progress.TimeElapsedColumn(),
            ) as progress:
                progress_task_id = progress.add_task("[red]Downloading Lavalink.jar", total=response.content_length)
                try:
                    chunk = await response.content.read(1024)
                    while chunk:
                        chunk_size = file.write(chunk)
                        nbytes += chunk_size
                        progress.update(progress_task_id, advance=chunk_size)
                        chunk = await response.content.read(1024)
                    file.flush()
                finally:
                    file.close()

            shutil.move(path, str(LAVALINK_JAR_FILE), copy_function=shutil.copyfile)

        LOGGER.info("Successfully downloaded Lavalink.jar (%s bytes written)", format(nbytes, ","))
        await self._is_up_to_date(forced=forced)
        # noinspection PyProtectedMember
        await self._client._config.update_download_id(self._ci_info["number"])

    async def _is_up_to_date(self, forced: bool = False) -> bool:
        if self._up_to_date is True and not forced:
            # Return cached value if we've checked this before
            return True
        # noinspection PyProtectedMember
        last_download_id = await self._client._config.fetch_download_id()
        args, _ = await self._get_jar_args()
        args.append("--version")
        _proc = await asyncio.subprocess.create_subprocess_exec(  # pylint:disable=no-member
            *args,
            cwd=str(LAVALINK_DOWNLOAD_DIR),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        stdout = (await _proc.communicate())[0]
        build = LAVALINK_BUILD_LINE.search(stdout) or {"build": b"Unknown"}
        branch = LAVALINK_BRANCH_LINE.search(stdout) or {"branch": b"Unknown"}
        java = LAVALINK_JAVA_LINE.search(stdout) or {"jvm": b"Unknown"}
        lavaplayer = LAVALINK_LAVAPLAYER_LINE.search(stdout) or {"lavaplayer": b"Unknown"}
        buildtime = LAVALINK_BUILD_TIME_LINE.search(stdout) or {"build_time": b"Unknown"}
        commit = LAVALINK_COMMIT_LINE.search(stdout) or {"commit": b"Unknown"}
        version = LAVALINK_VERSION_LINE.search(stdout) or {"version": b"Unknown"}
        LOGGER.info(
            "Current Lavalink meta: Lavalink build: %s, branch: %s, "
            "java: %s, lavaplayer: %s, build_time: %s commit: %s version: %s",
            build["build"],
            branch["branch"],
            java["jvm"],
            lavaplayer["lavaplayer"],
            buildtime["build_time"],
            commit["commit"],
            version["version"],
        )
        if build["build"] == b"Unknown":
            build = int(last_download_id)
        else:
            build = int(build["build"])
        date = buildtime["build_time"].decode()
        date = date.replace(".", "/")
        self._lavalink_build = build
        self._lavalink_branch = branch["branch"].decode()
        self._jvm = java["jvm"].decode()
        self._lavaplayer = lavaplayer["lavaplayer"].decode()
        self._commit = commit["commit"].decode()
        self._version = version["version"].decode()
        self._buildtime = date
        if await self.should_auto_update() or forced:
            self._up_to_date = last_download_id == self._ci_info.get("number", -1)
        else:
            self._ci_info["number"] = build
            self._up_to_date = True
        return self._up_to_date

    async def maybe_download_jar(self) -> None:
        if USING_FORCED is False:
            self._ci_info = await self.get_ci_latest_info()
        LOGGER.info("CI info: %s", self._ci_info)
        if not (await LAVALINK_JAR_FILE.exists() and await self._is_up_to_date()):
            await self._download_jar()

    async def wait_until_ready(self, timeout: float | None = None) -> None:
        tasks = [asyncio.create_task(c) for c in [self.ready.wait(), self.abort_for_unmanaged.wait()]]
        done, pending = await asyncio.wait(tasks, timeout=timeout or self.timeout, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        if done:
            done.pop().result()
        if self.abort_for_unmanaged.is_set():
            raise asyncio.TimeoutError
        if not self.ready.is_set():
            raise asyncio.TimeoutError

    async def wait_until_connected(self, timeout: float | None = None) -> None:
        tasks = [asyncio.create_task(c) for c in [self._wait_for.wait(), self.wait_until_ready()]]
        done, pending = await asyncio.wait(tasks, timeout=timeout or self.timeout, return_when=asyncio.ALL_COMPLETED)
        for task in pending:
            task.cancel()

    async def start_monitor(self, java_path: str) -> None:
        retry_count = 0
        backoff = ExponentialBackoffWithReset(base=3)
        while True:
            try:
                await self._monitor_primary_healthy_flow(backoff, java_path, retry_count)
            except (TooManyProcessFoundException, IncorrectProcessFoundException, NoProcessFoundException):
                await self._partial_shutdown()
            except asyncio.TimeoutError:
                delay = backoff.delay()
                await self._partial_shutdown()
                LOGGER.warning(
                    "Lavalink Managed node health check timeout, restarting in %s seconds",
                    delay,
                )
                await asyncio.sleep(delay)
            except ManagedLavalinkStartFailureException:
                await self._monitor_managed_start_failure()
            except NodeUnhealthyException:
                await self._monitor_handle_unhealthy(backoff)
            except LavalinkDownloadFailedException as exc:
                delay = backoff.delay()
                if exc.should_retry:
                    LOGGER.warning(
                        "Lavalink Managed node download failed retrying in %s seconds\n%s",
                        delay,
                        exc.response,
                    )
                    retry_count += 1
                    await self._partial_shutdown()
                    await asyncio.sleep(delay)
                else:
                    LOGGER.error(
                        "Fatal exception whilst starting managed Lavalink node, aborting...\n%s",
                        exc.response,
                    )
                    # lavalink_connection_aborted
                    return await self.shutdown()
            except InvalidArchitectureException:
                LOGGER.error("Invalid machine architecture, cannot run a managed Lavalink node")
                # lavalink_connection_aborted
                return await self.shutdown()
            except (UnsupportedJavaException, UnexpectedJavaResponseException) as exc:
                LOGGER.error(exc)
                # lavalink_connection_aborted
                return await self.shutdown()
            except ManagedLinkStartAbortedUseExternal:
                LOGGER.warning("Lavalink Managed node start aborted, using the detected external Lavalink node")
                return await self.connect_node(reconnect=False, wait_for=0, external_fallback=True)
            except ManagedLavalinkNodeException as exc:
                await self._monitor_managed_node_error(backoff, exc)
            except asyncio.CancelledError:
                LOGGER.warning("Lavalink Managed monitor task cancelled")
                return
            except Exception as exc:
                delay = backoff.delay()
                LOGGER.warning("Lavalink Managed node startup failed retrying in %s seconds", delay, exc_info=exc)
                await self._partial_shutdown()
                await asyncio.sleep(delay)

    async def _monitor_primary_healthy_flow(
        self, backoff: ExponentialBackoffWithReset, java_path: str, retry_count: int
    ) -> None:
        self._shutdown = False
        if self._node_pid is None or not psutil.pid_exists(self._node_pid):
            self.ready.clear()
            await self._start(java_path=java_path)
        while True:
            await self._monitor_primary_node_iteration(backoff, retry_count)

    async def _monitor_primary_node_iteration(self, backoff: ExponentialBackoffWithReset, retry_count: int) -> None:
        await self.wait_until_ready(timeout=self.timeout)
        if not psutil.pid_exists(self._node_pid):
            raise NoProcessFoundException
        if self._node is None or not self._node.websocket.connected and not self._node.websocket.connecting:
            await self.connect_node(reconnect=retry_count != 0, wait_for=3)
        try:
            await self._monitor_connect_to_node(backoff)
        except AttributeError as e:
            await self._monitor_wait_for_connection(e)
        except Exception as exc:
            LOGGER.debug(exc, exc_info=exc)
            raise NodeUnhealthyException(str(exc)) from exc

    async def _monitor_managed_node_error(self, backoff: ExponentialBackoffWithReset, exc: Exception) -> None:
        delay = backoff.delay()
        LOGGER.error(
            exc,
        )
        await self._partial_shutdown()
        LOGGER.warning("Lavalink Managed node startup failed retrying in %s seconds", delay, exc_info=exc)
        await asyncio.sleep(delay)

    async def _monitor_handle_unhealthy(self, backoff: ExponentialBackoffWithReset) -> None:
        delay = backoff.delay()
        await self._partial_shutdown()
        LOGGER.warning(
            "Lavalink Managed node health check failed, restarting in %s seconds",
            delay,
        )
        await asyncio.sleep(delay)

    async def _monitor_managed_start_failure(self) -> None:
        LOGGER.warning("Lavalink Managed node failed to start, restarting")
        await self._partial_shutdown()
        for process in iter(
            await self.get_lavalink_process(
                "-Djdk.tls.client.protocols=TLSv1.2", "-Xms64M", "-jar", cwd=str(LAVALINK_DOWNLOAD_DIR)
            )
        ):
            with contextlib.suppress(psutil.Error):
                pid = process["pid"]
                p = psutil.Process(pid)
                p.terminate()
                p.kill()

    async def _monitor_wait_for_connection(self, e: Exception) -> None:
        try:
            LOGGER.debug("Managed node monitor detected PyLav is not connected to any nodes -%s", exc_info=e)
            while True:
                node = self._client.node_manager.get_node_by_id(self._node_id)
                if node is not None:
                    await node.wait_until_ready(timeout=30)
                if node and node.websocket.connected:
                    break
                await asyncio.sleep(1)
        except asyncio.TimeoutError:
            raise

    async def _monitor_connect_to_node(self, backoff: ExponentialBackoffWithReset) -> None:
        node = self._client.node_manager.get_node_by_id(self._node_id)
        if node is not None:
            await node.wait_until_ready(timeout=30)
        if node.websocket.connected:
            try:
                # Hoping this throws an exception which will then trigger a restart
                await node.websocket.ping()
                backoff.reset()
                await asyncio.sleep(1)
            except WebsocketNotConnectedException:
                await asyncio.sleep(5)
            except ConnectionResetError as e:
                raise AttributeError from e
        elif node.websocket.connecting:
            await node.websocket.wait_until_ready(timeout=30)
        else:
            raise AttributeError

    async def start(self, java_path: str) -> None:
        self._disabled = False
        self._java_path = java_path
        if self.start_monitor_task is not None:
            await self.shutdown()
            self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=json.dumps)
        if self.__buffer_task is not None:
            self.__buffer_task.cancel()
            self.__buffer_task = None
        self._wait_for.clear()

        await update_plugins(self._client)
        self.start_monitor_task = asyncio.create_task(self.start_monitor(java_path))
        self.start_monitor_task.set_name("LavalinkManagedNode.health_monitor")

    async def connect_node(self, reconnect: bool, wait_for: float = 0.0, external_fallback: bool = False) -> None:
        # sourcery no-metrics
        await asyncio.sleep(wait_for)
        self._wait_for.clear()
        if not self.ready.is_set():
            if external_fallback:
                self.ready.set()
            else:
                raise ManagedLavalinkStartFailureException()
        if reconnect:
            node = self._client.node_manager.get_node_by_id(self._node_id)
            if node is not None:
                self._node = node
                if node.websocket.connecting:
                    await node.wait_until_ready(timeout=30)
                elif node.websocket.connected:
                    LOGGER.info("Managed Lavalink node is connected")
                else:
                    LOGGER.info("Managed Lavalink node is not connected, reconnecting")
                    await self.restart()
                    return
                self._wait_for.set()
                return
        if (node := self._client.node_manager.get_node_by_id(self._node_id)) is None:
            node = await self.connect_to_node(external_fallback)
        else:
            self._node = node
        if node.websocket.connecting:
            await node.wait_until_ready()
        elif node.websocket.connected:
            LOGGER.info("Managed Lavalink node is connected")
        else:
            LOGGER.info("Managed Lavalink node is not connected, reconnecting")
            await node.websocket.close()
            # noinspection PyProtectedMember
            await node.websocket._websocket_closed(reason="Managed Node restart")
            await node.wait_until_ready(timeout=60)
        self._wait_for.set()

    async def connect_to_node(self, external_fallback: bool) -> Node:
        data = await self._client.node_db_manager.bundled_node_config().fetch_all()
        name = (
            f"PyLavPortConflictRecovery: {self._node_pid}"
            if external_fallback
            else f"PyLavManagedNode: {self._node_pid}"
        )
        data["yaml"]["sentry"]["tags"]["pylav_version"] = self._client.lib_version
        node = self._node = await self._client.add_node(
            host=self._current_config["server"]["address"],
            port=self._current_config["server"]["port"],
            password=self._current_config["lavalink"]["server"]["password"],
            resume_timeout=data["resume_timeout"],
            name=name,
            managed=True,
            ssl=False,
            search_only=False,
            unique_identifier=self._client.node_db_manager.bundled_node_config().id,
            temporary=True,
        )
        await node.config.update_name(name)
        return node

    @staticmethod
    async def get_lavalink_process(
        *matches: str, cwd: str | None = None, lazy_match: bool = False
    ) -> list[dict[str, Any]]:
        process_list = []
        filter_ = [cwd] if cwd else []
        for proc in psutil.process_iter():
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                if cwd and await asyncio.to_thread(proc.cwd) not in filter_:
                    continue
                cmdline = await asyncio.to_thread(proc.cmdline)
                if (
                    matches
                    and all(a in cmdline for a in iter(matches))
                    or lazy_match
                    and any("lavalink" in arg.lower() for arg in iter(cmdline))
                ):
                    proc_as_dict = await asyncio.to_thread(
                        proc.as_dict, attrs=["pid", "name", "create_time", "status", "cmdline", "cwd"]
                    )

                    process_list.append(proc_as_dict)
        return process_list

    async def restart(self, java_path: str = None) -> None:
        LOGGER.info("Restarting managed Lavalink node")
        if node := self._client.get_my_node():
            if self.start_monitor_task is not None:
                self.start_monitor_task.cancel()
                self.start_monitor_task = None
            if self.__buffer_task is not None:
                self.__buffer_task.cancel()
                self.__buffer_task = None

            # noinspection PyProtectedMember
            if not node.websocket._manual_shutdown:
                await node.websocket.manual_closure(managed_node=True)
            with contextlib.suppress(Exception):
                await node.close()
        await self.shutdown()
        await self.start(java_path=java_path or self._java_path)



from __future__ import annotations

import asyncio
import functools
import threading
from collections.abc import Awaitable, Callable
from typing import Any

from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE

_LOCK_SINGLETON_CLASS = threading.Lock()
_LOCK_SINGLETON_CALLABLE = threading.Lock()
_LOCK_SINGLETON_CACHE_CALLABLE = threading.Lock()


def synchronized_method_call(
    lock: threading.Lock, discard: bool = False
) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
    """Synchronization decorator"""

    def wrapper(
        f: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
        @functools.wraps(f)
        def inner_wrapper(
            *args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs
        ) -> ANY_GENERIC_TYPE | Awaitable[ANY_GENERIC_TYPE] | None:
            if lock.locked() and discard:
                return
            with lock:
                return f(*args, **kwargs)

        return inner_wrapper

    return wrapper


def synchronized_method_call_with_self_threading_lock(
    discard: bool = False,
) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
    """Synchronization decorator"""

    def wrapper(
        f: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
        @functools.wraps(f)
        def inner_wrapper(
            self, *args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs
        ) -> ANY_GENERIC_TYPE | Awaitable[ANY_GENERIC_TYPE] | None:
            if self._threading_lock.locked() and discard:
                return
            with self._threading_lock:
                return f(self, *args, **kwargs)

        return inner_wrapper

    return wrapper


class SingletonClass(type):
    _instances = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> SingletonClass:
        # sourcery skip: instance-method-first-arg-name
        if cls not in cls._instances:
            cls._locked_call(*args, **kwargs)
        return cls._instances[cls]

    @synchronized_method_call(_LOCK_SINGLETON_CLASS)
    def _locked_call(cls, *args: Any, **kwargs: Any) -> None:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)


class SingletonCallable:
    _has_run = {}
    _responses = {}

    @classmethod
    @synchronized_method_call(_LOCK_SINGLETON_CALLABLE)
    def reset(cls) -> None:
        cls._has_run = {}
        cls._responses = {}

    @classmethod
    @synchronized_method_call(_LOCK_SINGLETON_CALLABLE)
    def run_once(
        cls, f: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
        @functools.wraps(f)
        def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> ANY_GENERIC_TYPE:
            if not cls._has_run.get(f, False):
                cls._has_run[f] = True
                output = f(*args, **kwargs)
                cls._responses[f] = output
                return output
            else:
                return cls._responses.get(f, None)

        cls._has_run[f] = False
        return wrapper

    @classmethod
    @synchronized_method_call(_LOCK_SINGLETON_CALLABLE)
    def run_once_async(
        cls, f: Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]  # type: ignore
    ) -> Callable[PARAM_SPEC_TYPE, Awaitable[ANY_GENERIC_TYPE]]:  # type: ignore
        @functools.wraps(f)
        def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> Awaitable[ANY_GENERIC_TYPE]:
            if not cls._has_run.get(f, False):
                cls._has_run[f.__name__] = True
                return f(*args, **kwargs)
            else:
                return asyncio.sleep(0)

        cls._has_run[f] = False
        return wrapper


class SingletonCachedByKey(type):
    _instances: Any = {}

    @classmethod
    def _get_key(cls, mro, **kwargs: Any) -> tuple[str, ...] | None:
        singleton_key = f'{kwargs.get("id")}'
        for base in mro:
            if base.__module__.startswith("pylav.storage.models"):
                key_name = base.__name__
                if key_name in ("PlayerState", "NodeMock", "Equalizer"):
                    return
                if key_name in ("PlayerConfig", "Config"):
                    singleton_key += f".{kwargs.get('bot')}"
                return singleton_key, key_name

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        # sourcery skip: instance-method-first-arg-name
        key = cls._get_key(cls.mro(), **kwargs)
        if key not in cls._instances:
            rps = cls._locked_call(*args, **kwargs)
            if key is None:
                return rps
        return cls._instances[key]

    @synchronized_method_call(_LOCK_SINGLETON_CACHE_CALLABLE)
    def _locked_call(cls, *args: Any, **kwargs: Any) -> None:
        key = cls._get_key(cls.mro(), **kwargs)
        if key not in cls._instances:
            singleton = super().__call__(*args, **kwargs)
            if key is not None:
                cls._instances[key] = singleton
            else:
                return singleton



from __future__ import annotations

import typing
from dataclasses import dataclass

import discord

from pylav.compat import json
from pylav.helpers.misc import TimedFeature
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.players import PlayerRow
from pylav.type_hints.bot import DISCORD_BOT_TYPE
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class PlayerConfig(CachedModel, metaclass=SingletonCachedByKey):
    id: int
    bot: int

    def get_cache_key(self) -> str:
        return f"{self.id}:{self.bot}"

    @classmethod
    async def create_global(cls, bot: int) -> None:
        """Create the player in the database"""
        data = json.dumps(
            {
                "enabled": False,
                "time": 60,
            }
        )
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """
            INSERT INTO player
            (id, bot, volume,
            max_volume, shuffle, auto_shuffle,
            auto_play, self_deaf, empty_queue_dc,
            alone_dc, alone_pause)
            VALUES ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})
            ON CONFLICT (id, bot) DO NOTHING;
            ;
            """,
            0,
            bot,
            1000,
            1000,
            True,
            True,
            True,
            True,
            data,
            data,
            data,
        )

    async def delete(self) -> None:
        """Delete the player from the database"""
        await PlayerRow.delete().where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
        await self.invalidate_cache()

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_TYPE:
        """Get all players from the database"""
        data = (
            await PlayerRow.select()
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        if data:
            del data["primary_key"]
            data["empty_queue_dc"] = TimedFeature.from_dict(data["empty_queue_dc"])
            data["alone_dc"] = TimedFeature.from_dict(data["alone_dc"])
            data["alone_pause"] = TimedFeature.from_dict(data["alone_pause"])
            data["extras"] = data["extras"]
            data["effects"] = data["effects"]
            return data
        return {
            "id": self.id,
            "bot": self.bot,
            "volume": PlayerRow.volume.default,
            "max_volume": PlayerRow.max_volume.default,
            "auto_play_playlist_id": PlayerRow.auto_play_playlist_id.default,
            "text_channel_id": PlayerRow.text_channel_id.default,
            "notify_channel_id": PlayerRow.notify_channel_id.default,
            "forced_channel_id": PlayerRow.forced_channel_id.default,
            "repeat_current": PlayerRow.repeat_current.default,
            "repeat_queue": PlayerRow.repeat_queue.default,
            "shuffle": PlayerRow.shuffle.default,
            "auto_shuffle": PlayerRow.auto_shuffle.default,
            "auto_play": PlayerRow.auto_play.default,
            "self_deaf": PlayerRow.self_deaf.default,
            "empty_queue_dc": TimedFeature.from_dict(json.loads(PlayerRow.empty_queue_dc.default)),
            "alone_dc": TimedFeature.from_dict(json.loads(PlayerRow.alone_dc.default)),
            "alone_pause": TimedFeature.from_dict(json.loads(PlayerRow.alone_pause.default)),
            "extras": json.loads(PlayerRow.extras.default),
            "effects": json.loads(PlayerRow.effects.default),
            "dj_users": PlayerRow.dj_users.default,
            "dj_roles": PlayerRow.dj_roles.default,
        }

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the player exists in the database"""
        return await PlayerRow.exists().where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))

    @maybe_cached
    async def fetch_volume(self) -> int:
        """Fetch the volume of the player from the db"""

        player = (
            await PlayerRow.select(PlayerRow.volume)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["volume"] if player else PlayerRow.volume.default

    async def update_volume(self, volume: int) -> None:
        """Update the volume of the player in the db"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, volume)
        VALUES ({}, {}, {})
        ON CONFLICT (id, bot)
         DO UPDATE SET volume = excluded.volume;""",
            self.id,
            self.bot,
            volume,
        )
        await self.update_cache((self.fetch_volume, volume), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_max_volume(self) -> int:
        """Fetch the max volume of the player from the db"""
        player = (
            await PlayerRow.select(PlayerRow.max_volume)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["max_volume"] if player else PlayerRow.max_volume.default

    async def update_max_volume(self, max_volume: int) -> None:
        """Update the max volume of the player in the db"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, max_volume)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET max_volume = excluded.max_volume;""",
            self.id,
            self.bot,
            max_volume,
        )
        await self.update_cache((self.fetch_max_volume, max_volume), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_play_playlist_id(self) -> int:
        """Fetch the auto play playlist ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_play_playlist_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_play_playlist_id"] if player else PlayerRow.auto_play_playlist_id.default

    async def update_auto_play_playlist_id(self, auto_play_playlist_id: int) -> None:
        """Update the auto play playlist ID of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, auto_play_playlist_id)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET auto_play_playlist_id = excluded.auto_play_playlist_id;""",
            self.id,
            self.bot,
            auto_play_playlist_id,
        )
        await self.update_cache((self.fetch_auto_play_playlist_id, auto_play_playlist_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_text_channel_id(self) -> int:
        """Fetch the text channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.text_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["text_channel_id"] if player else PlayerRow.text_channel_id.default

    async def update_text_channel_id(self, text_channel_id: int) -> None:
        """Update the text channel ID of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, text_channel_id)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET text_channel_id = excluded.text_channel_id;""",
            self.id,
            self.bot,
            text_channel_id,
        )
        await self.update_cache((self.fetch_text_channel_id, text_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_notify_channel_id(self) -> int:
        """Fetch the notify channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.notify_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["notify_channel_id"] if player else PlayerRow.notify_channel_id.default

    async def update_notify_channel_id(self, notify_channel_id: int) -> None:
        """Update the notify channel ID of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, notify_channel_id)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET notify_channel_id = excluded.notify_channel_id;""",
            self.id,
            self.bot,
            notify_channel_id,
        )
        await self.update_cache((self.fetch_notify_channel_id, notify_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_forced_channel_id(self) -> int:
        """Fetch the forced channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.forced_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["forced_channel_id"] if player else PlayerRow.forced_channel_id.default

    async def update_forced_channel_id(self, forced_channel_id: int) -> None:
        """Update the forced channel ID of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, forced_channel_id)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET forced_channel_id = excluded.forced_channel_id;""",
            self.id,
            self.bot,
            forced_channel_id,
        )
        await self.update_cache((self.fetch_forced_channel_id, forced_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_repeat_current(self) -> bool:
        """Fetch the repeat current of the player"""
        player = (
            await PlayerRow.select(PlayerRow.repeat_current)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["repeat_current"] if player else PlayerRow.repeat_current.default

    async def update_repeat_current(self, repeat_current: bool) -> None:
        """Update the repeat current of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, repeat_current)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET repeat_current = excluded.repeat_current;""",
            self.id,
            self.bot,
            repeat_current,
        )
        await self.update_cache((self.fetch_repeat_current, repeat_current), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_repeat_queue(self) -> bool:
        """Fetch the repeat queue of the player"""
        player = (
            await PlayerRow.select(PlayerRow.repeat_queue)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["repeat_queue"] if player else PlayerRow.repeat_queue.default

    async def update_repeat_queue(self, repeat_queue: bool) -> None:
        """Update the repeat queue of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, repeat_queue)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET repeat_queue = excluded.repeat_queue;""",
            self.id,
            self.bot,
            repeat_queue,
        )
        await self.update_cache((self.fetch_repeat_queue, repeat_queue), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_shuffle(self) -> bool:
        """Fetch the shuffle of the player"""
        player = (
            await PlayerRow.select(PlayerRow.shuffle)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["shuffle"] if player else PlayerRow.shuffle.default

    async def update_shuffle(self, shuffle: bool) -> None:
        """Update the shuffle of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, shuffle)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET shuffle = excluded.shuffle;""",
            self.id,
            self.bot,
            shuffle,
        )
        await self.update_cache((self.fetch_shuffle, shuffle), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_shuffle(self) -> bool:
        """Fetch the auto shuffle of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_shuffle)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_shuffle"] if player else PlayerRow.auto_shuffle.default

    async def update_auto_shuffle(self, auto_shuffle: bool) -> None:
        """Update the auto shuffle of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, auto_shuffle)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET auto_shuffle = excluded.auto_shuffle;""",
            self.id,
            self.bot,
            auto_shuffle,
        )
        await self.update_cache((self.fetch_auto_shuffle, auto_shuffle), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_play(self) -> bool:
        """Fetch the auto play of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_play)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_play"] if player else PlayerRow.auto_play.default

    async def update_auto_play(self, auto_play: bool) -> None:
        """Update the auto play of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, auto_play)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET auto_play = excluded.auto_play;""",
            self.id,
            self.bot,
            auto_play,
        )
        await self.update_cache((self.fetch_auto_play, auto_play), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_self_deaf(self) -> bool:
        """Fetch the self deaf of the player"""
        player = (
            await PlayerRow.select(PlayerRow.self_deaf)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["self_deaf"] if player else PlayerRow.self_deaf.default

    async def update_self_deaf(self, self_deaf: bool) -> None:
        """Update the self deaf of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, self_deaf)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET self_deaf = excluded.self_deaf;""",
            self.id,
            self.bot,
            self_deaf,
        )
        await self.update_cache((self.fetch_self_deaf, self_deaf), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the extras of the player"""
        player = (
            await PlayerRow.select(PlayerRow.extras)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["extras"] if player else json.loads(PlayerRow.extras.default)

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the extras of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, extras)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET extras = excluded.extras;""",
            self.id,
            self.bot,
            json.dumps(extras),
        )
        await self.update_cache((self.fetch_extras, extras), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_effects(
        self,
    ) -> dict[str, int | None | dict[str, int | float | list[dict[str, float | None]] | None]]:
        """Fetch the effects of the player"""
        player = (
            await PlayerRow.select(PlayerRow.effects)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["effects"] if player else json.loads(PlayerRow.effects.default)

    async def update_effects(
        self, effects: dict[str, int | None | dict[str, int | float | list[dict[str, float | None]] | None]]
    ) -> None:
        """Update the effects of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, effects)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET effects = excluded.effects;""",
            self.id,
            self.bot,
            json.dumps(effects),
        )
        await self.update_cache((self.fetch_effects, effects), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_empty_queue_dc(self) -> TimedFeature:
        """Fetch the empty queue dc of the player"""
        player = (
            await PlayerRow.select(PlayerRow.empty_queue_dc)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(
            player["empty_queue_dc"] if player else json.loads(PlayerRow.empty_queue_dc.default)
        )

    async def update_empty_queue_dc(self, empty_queue_dc: dict[str, int | bool]) -> None:
        """Update the empty queue dc of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, empty_queue_dc)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET empty_queue_dc = excluded.empty_queue_dc;""",
            self.id,
            self.bot,
            json.dumps(empty_queue_dc),
        )
        await self.update_cache(
            (self.fetch_empty_queue_dc, TimedFeature.from_dict(empty_queue_dc)), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_alone_dc(self) -> TimedFeature:
        """Fetch the alone dc of the player"""
        player = (
            await PlayerRow.select(PlayerRow.alone_dc)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(player["alone_dc"] if player else json.loads(PlayerRow.alone_dc.default))

    async def update_alone_dc(self, alone_dc: dict[str, int | bool]) -> None:
        """Update the alone dc of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, alone_dc)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET alone_dc = excluded.alone_dc;""",
            self.id,
            self.bot,
            json.dumps(alone_dc),
        )
        await self.update_cache((self.fetch_alone_dc, TimedFeature.from_dict(alone_dc)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_alone_pause(self) -> TimedFeature:
        """Fetch the alone pause of the player"""
        player = (
            await PlayerRow.select(PlayerRow.alone_pause)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(player["alone_pause"] if player else json.loads(PlayerRow.alone_pause.default))

    async def update_alone_pause(self, alone_pause: dict[str, int | bool]) -> None:
        """Update the alone pause of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, alone_pause)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET alone_pause = excluded.alone_pause;""",
            self.id,
            self.bot,
            json.dumps(alone_pause),
        )
        await self.update_cache((self.fetch_alone_pause, TimedFeature.from_dict(alone_pause)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_dj_users(self) -> set[int]:
        """Fetch the disc jockey users of the player"""
        player = (
            await PlayerRow.select(PlayerRow.dj_users)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )

        return set(player["dj_users"] if player else PlayerRow.dj_users.default)

    async def add_to_dj_users(self, user: discord.Member) -> None:
        """Add a user to the disc jockey users of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, dj_users)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET dj_users = array_cat(player.dj_users, EXCLUDED.dj_users);""",
            self.id,
            self.bot,
            [user.id],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def remove_from_dj_users(self, user: discord.Member) -> None:
        """Remove a user from the disc jockey users of the player"""
        # TODO: When piccolo add more functions for dealing with arrays update this to become ORM
        await PlayerRow.raw(
            "UPDATE player SET dj_users = array_remove(dj_users, {}) WHERE id = {} AND bot = {};",
            user.id,
            self.id,
            self.bot,
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def bulk_add_dj_users(self, *users: discord.Member) -> None:
        """Add disc jockey users to the player"""
        if not users:
            return
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, dj_users)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET dj_users = array_cat(player.dj_users, EXCLUDED.dj_users);""",
            self.id,
            self.bot,
            [u.id for u in users],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def bulk_remove_dj_users(self, *users: discord.Member) -> None:
        """Remove disc jockey users from the player.

        Parameters
        ----------
        users : discord.Member
            The users to add
        """
        if not users:
            return
        for user in users:
            await self.remove_from_dj_users(user)

    async def dj_users_reset(self) -> None:
        """Reset the disc jockey users of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """
            INSERT INTO player (id, bot, dj_users) VALUES ({}, {}, {})
            ON CONFLICT (id, bot) DO UPDATE SET dj_users = excluded.dj_users;
            """,
            self.id,
            self.bot,
            [],
        )
        await self.update_cache((self.fetch_dj_users, set()), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_dj_roles(self) -> set[int]:
        """Fetch the disc jockey roles of the player"""
        player = (
            await PlayerRow.select(PlayerRow.dj_roles)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )

        return set(player["dj_roles"] if player else PlayerRow.dj_roles.default)

    async def add_to_dj_roles(self, role: discord.Role) -> None:
        """Add disc jockey roles to the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, dj_roles)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET dj_roles = array_cat(player.dj_roles, EXCLUDED.dj_roles)""",
            self.id,
            self.bot,
            [role.id],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def remove_from_dj_roles(self, role: discord.Role) -> None:
        """Remove disc jockey roles from the player"""
        # TODO: When piccolo add more functions for dealing with arrays update this to become ORM

        await PlayerRow.raw(
            """UPDATE player SET dj_roles = array_remove(dj_roles, {}) WHERE id = {} AND bot = {}""",
            role.id,
            self.id,
            self.bot,
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def bulk_add_dj_roles(self, *roles: discord.Role) -> None:
        """Add disc jockey roles to the player.

        Parameters
        ----------
        roles : discord.Role
            The roles to add"
        """
        if not roles:
            return
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """INSERT INTO player (id, bot, dj_roles)
                    VALUES ({}, {}, {})
                    ON CONFLICT (id, bot)
                    DO UPDATE SET dj_roles = array_cat(player.dj_roles, EXCLUDED.dj_roles);""",
            self.id,
            self.bot,
            [r.id for r in roles],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def bulk_remove_dj_roles(self, *roles: discord.Role) -> None:
        """Remove disc jockey roles from the player.

        Parameters
        ----------
        roles : discord.Role
            The roles to add.
        """
        if not roles:
            return
        for role in roles:
            await self.remove_from_dj_roles(role)

    async def dj_roles_reset(self) -> None:
        """Reset the disc jockey roles of the player"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await PlayerRow.raw(
            """
                    INSERT INTO player (id, bot, dj_roles) VALUES ({}, {}, {})
                    ON CONFLICT (id, bot) DO UPDATE SET dj_roles = excluded.dj_roles;
                    """,
            self.id,
            self.bot,
            [],
        )
        await self.update_cache((self.fetch_dj_roles, set()), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    async def _role_id_in_dj_roles(self, role_id: int) -> bool:
        return await PlayerRow.exists().where(
            (PlayerRow.id == self.id) & (PlayerRow.bot == self.bot) & PlayerRow.dj_roles.any(role_id)
        )

    async def _userid_in_dj_users(self, user_id: int) -> bool:
        return await PlayerRow.exists().where(
            (PlayerRow.id == self.id) & (PlayerRow.bot == self.bot) & PlayerRow.dj_users.any(user_id)
        )

    async def is_dj(
        self,
        user: discord.Member,
        *,
        additional_role_ids: list | None = None,
        additional_user_ids: list | None = None,
        bot: DISCORD_BOT_TYPE = None,
    ) -> bool:
        """Check if a user is a disc jockey.

        Parameters
        ----------
        user : discord.Member
            The user to check.
        additional_role_ids : list
            The additional disc jockey role ids to check.
        additional_user_ids : list
            The additional disc jockey user ids to check.
        bot : DISCORD_BOT_TYPE
            The bot instance to check for owners, admins or mods.

        Returns
        -------
        bool
            Whether the user is a disc jockey.
        """
        if additional_user_ids and user.id in additional_user_ids:
            return True
        if additional_role_ids and any(r.id in additional_role_ids for r in user.roles):
            return True
        if __ := user.guild:
            if hasattr(bot, "is_owner") and await bot.is_owner(typing.cast(discord.User, user)):
                return True
            if hasattr(bot, "is_admin") and await bot.is_admin(user):
                return True
            if hasattr(bot, "is_mod") and await bot.is_mod(user):
                return True
        if await self._userid_in_dj_users(user.id):
            return True
        dj_roles = await self.fetch_dj_roles()
        return bool(any(r.id in dj_roles for r in user.roles)) if dj_roles else True



from __future__ import annotations

from collections.abc import Awaitable, Callable, Coroutine
from typing import Any, ParamSpec, Protocol, TypeVar, Union

ANY_GENERIC_TYPE = TypeVar("ANY_GENERIC_TYPE")
PARAM_SPEC_TYPE = ParamSpec("PARAM_SPEC_TYPE")

CORO_TYPE = Coroutine[Any, Any, ANY_GENERIC_TYPE]
CORO_FUNCTION = Callable[..., CORO_TYPE[Any]]
MaybeCoro = Union[ANY_GENERIC_TYPE, CORO_TYPE[ANY_GENERIC_TYPE]]
MaybeAwaitable = Union[ANY_GENERIC_TYPE, Awaitable[ANY_GENERIC_TYPE]]


class SupportsStr(Protocol):
    def __str__(self) -> str:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.events.base import PyLavEvent
from pylav.events.track.track_start import TrackStartEvent as TrackStartEvent
from pylav.nodes.api.responses.websocket import TrackEnd, TrackException, TrackStuck

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class TrackStuckEvent(PyLavEvent):
    """This event is dispatched when the currently playing track is stuck.
    This normally has something to do with the stream you are playing
    and not Lavalink itself.
    """

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStuck):
        self.player = player
        self.track = track
        self.threshold = event_object.thresholdMs
        self.node = node
        self.event = event_object


class TrackExceptionEvent(PyLavEvent):
    """This event is dispatched when an exception occurs while playing a track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackException):
        self.player = player
        self.track = track
        self.exception = event_object.exception.cause
        self.node = node
        self.event = event_object


class TrackEndEvent(PyLavEvent):
    """This event is dispatched when the player finished playing a track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackEnd) -> None:
        self.player = player
        self.track = track
        self.reason = event_object.reason
        self.node = node
        self.event = event_object


class TrackAutoPlayEvent(PyLavEvent):
    """This event is dispatched when the player starts to play a track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track) -> None:
        self.player = player
        self.track = track


class TrackResumedEvent(PyLavEvent):
    """This event is dispatched when the player resumes playing a track."""

    __slots__ = ()

    def __init__(self, player: Player, track: Track, requester: discord.Member) -> None:
        self.player = player
        self.track = track
        self.requester = requester


class TrackSeekEvent(PyLavEvent):
    """This event is dispatched when the player is moved."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, track: Track, before: float, after: float) -> None:
        self.player = player
        self.requester = requester
        self.track = track
        self.before = before
        self.after = after


class TrackPreviousRequestedEvent(PyLavEvent):
    """This event is dispatched when a history track is requested."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, track: Track) -> None:
        self.player = player
        self.requester = requester
        self.track = track


class TracksRequestedEvent(PyLavEvent):
    """This event is dispatched when a track in added to the queue."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, tracks: list[Track]) -> None:
        self.player = player
        self.requester = requester
        self.tracks = tracks


class TrackSkippedEvent(PyLavEvent):
    """This event is dispatched when a track is skipped."""

    __slots__ = ()

    def __init__(self, player: Player, requester: discord.Member, track: Track, position: float) -> None:
        self.player = player
        self.track = track
        self.requester = requester
        self.position = position



from __future__ import annotations

import asyncio
import contextlib
import inspect
import threading
from pathlib import Path
from types import MethodType

import discord
from discord.ext.commands import CheckFailure
from redbot.core import commands
from redbot.core.data_manager import cog_data_path
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import box
from tabulate import tabulate

from pylav.core.client import Client
from pylav.core.context import PyLavContext
from pylav.exceptions.node import NoNodeAvailableException, NoNodeWithRequestFunctionalityAvailableException
from pylav.extension.red.errors import (
    IncompatibleException,
    MediaPlayerNotFoundError,
    NotDJError,
    UnauthorizedChannelError,
)
from pylav.helpers.format.ascii import EightBitANSI
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE

_ = Translator("PyLav", Path(__file__))
_LOCK = threading.Lock()
LOGGER = getLogger("PyLav.ext.red.utils.overrides")

INCOMPATIBLE_COGS = {}


@commands.command(
    cls=commands.commands._AlwaysAvailableCommand,
    name="plcredits",
    aliases=["pltranslation"],
    i18n=_,
)
async def pylav_credits(context: PyLavContext) -> None:
    """Shows the credits and translation details for the PyLav cogs and shared code"""
    await context.send(
        embed=await context.pylav.construct_embed(
            messageable=context,
            description=_(
                "PyLav was created by {library_author_name_variable_do_not_translate}.\n\n"
                "PyLav source code can be located in {library_url_variable_do_not_translate}\n"
                "PyLav-Cogs source code can be located in {second_library_url_variable_do_not_translate}\n\n"
                "You can join the PyLav support server via {support_server_url_variable_do_not_translate}\n"
                "\n\n"
                "You can help translate PyLav by contributing to our Crowdin project at:\n"
                "{crowdin_project_url_variable_do_not_translate}\n\n\n"
                "Contributors:\n"
                "- {project_contributors_url_variable_do_not_translate}\n"
                "- {second_project_contributors_url_variable_do_not_translate}\n"
                "If you wish to buy me a coffee for my work, you can do so at:\n"
                "{buymeacoffee_url_variable_do_not_translate} or {github_sponsors_url_variable_do_not_translate}"
            ).format(
                library_author_name_variable_do_not_translate="[Draper#6666](https://github.com/Drapersniper)",
                library_url_variable_do_not_translate="https://github.com/Drapersniper/PyLav",
                second_library_url_variable_do_not_translate="https://github.com/Drapersniper/PyLav-Cogs",
                support_server_url_variable_do_not_translate="https://discord.com/invite/Sjh2TSCYQB",
                crowdin_project_url_variable_do_not_translate="https://crowdin.com/project/pylav",
                project_contributors_url_variable_do_not_translate="https://github.com/Drapersniper/PyLav/graphs/contributors",
                second_project_contributors_url_variable_do_not_translate="https://github.com/Drapersniper/PyLav-Cogs/graphs/contributors",
                buymeacoffee_url_variable_do_not_translate="https://www.buymeacoffee.com/draper",
                github_sponsors_url_variable_do_not_translate="https://github.com/sponsors/Drapersniper",
            ),
        ),
        ephemeral=True,
    )


@commands.command(
    cls=commands.commands._AlwaysAvailableCommand,
    name="plversion",
    aliases=["pylavversion"],
    i18n=_,
)
async def pylav_version(context: PyLavContext) -> None:
    """Show the version of PyLav library"""
    if isinstance(context, discord.Interaction):
        context = await context.client.get_context(context)
    if context.interaction and not context.interaction.response.is_done():
        await context.defer(ephemeral=True)
    data = [
        (EightBitANSI.paint_white("PyLav"), EightBitANSI.paint_blue(context.pylav.lib_version)),
    ]

    await context.send(
        embed=await context.pylav.construct_embed(
            description=box(
                tabulate(
                    data,
                    headers=(
                        EightBitANSI.paint_yellow(_("Library"), bold=True, underline=True),
                        EightBitANSI.paint_yellow(_("Version"), bold=True, underline=True),
                    ),
                    tablefmt="fancy_grid",
                ),
                lang="ansi",
            ),
            messageable=context,
        ),
        ephemeral=True,
    )


@commands.command(
    name="plsyncslash",
    aliases=["plss"],
    i18n=_,
)
@commands.is_owner()
async def pylav_sync_slash(context: PyLavContext) -> None:
    """Sync the Bots slash commands"""
    if isinstance(context, discord.Interaction):
        context = await context.client.get_context(context)
    if context.interaction and not context.interaction.response.is_done():
        await context.defer(ephemeral=True)
    await context.bot.wait_until_ready()
    await context.bot.tree.sync()
    await context.send(
        embed=await context.pylav.construct_embed(
            description=_("I have synced all my slash commands."),
            messageable=context,
        ),
        ephemeral=True,
    )


def _done_callback(task: asyncio.Task) -> None:
    with contextlib.suppress(asyncio.CancelledError):
        exc = task.exception()
        if exc is not None:
            LOGGER.error("Error in initialize task", exc_info=exc)


async def cog_command_error(self: DISCORD_COG_TYPE, context: PyLavContext, error: Exception) -> None:
    error = getattr(error, "original", error)
    unhandled = True
    if isinstance(error, MediaPlayerNotFoundError):
        unhandled = False
        await context.send(
            embed=await self.pylav.construct_embed(
                messageable=context,
                description=_("This command requires that I be in a voice channel before it can be executed."),
            ),
            ephemeral=True,
        )
    elif isinstance(error, NoNodeAvailableException):
        unhandled = False
        await context.send(
            embed=await self.pylav.construct_embed(
                messageable=context,
                description=_(
                    "PyLavPlayer cog is temporarily unavailable due to an outage with the backend services; please try again later."
                ),
                footer=_("There are no nodes available currently.")
                if await self.bot.is_owner(context.author)
                else None,
            ),
            ephemeral=True,
        )
    elif isinstance(error, NoNodeWithRequestFunctionalityAvailableException):
        unhandled = False
        await context.send(
            embed=await self.pylav.construct_embed(
                messageable=context,
                description=_(
                    "PyLavPlayer is currently unable to process tracks belonging to {feature_name_variable_do_not_translate}."
                ).format(feature_name_variable_do_not_translate=error.feature),
                footer=_(
                    "There is currently no available Lavalink node with the feature {feature_name_variable_do_not_translate}."
                ).format(feature_name_variable_do_not_translate=error.feature)
                if await self.bot.is_owner(context.author)
                else None,
            ),
            ephemeral=True,
        )
    elif isinstance(error, UnauthorizedChannelError):
        unhandled = False
        await context.send(
            embed=await self.pylav.construct_embed(
                messageable=context,
                description=_(
                    "This command is unavailable in this channel; please use {channel_name_variable_do_not_translate} instead."
                ).format(
                    channel_name_variable_do_not_translate=channel.mention
                    if (channel := context.guild.get_channel_or_thread(error.channel))
                    else error.channel
                ),
            ),
            ephemeral=True,
            delete_after=10,
        )
    elif isinstance(error, NotDJError):
        unhandled = False
        await context.send(
            embed=await self.pylav.construct_embed(
                messageable=context,
                description=_("This command requires you to be a disc jockey."),
            ),
            ephemeral=True,
            delete_after=10,
        )
    if unhandled:
        if (meth := getattr(self, "__pylav_original_cog_command_error", None)) and (
            func := self._get_overridden_method(meth)
        ):
            return await discord.utils.maybe_coroutine(func, context, error)
        else:
            return await self.bot.on_command_error(context, error, unhandled_by_cog=True)  # type: ignore


async def cog_unload(self: DISCORD_COG_TYPE) -> None:
    if self._init_task is not None:
        self._init_task.cancel()
    client = self.pylav
    await client.unregister(cog=self)
    if client._shutting_down:
        self.bot.remove_command(pylav_credits.qualified_name)
        self.bot.remove_command(pylav_version.qualified_name)
    if meth := getattr(self, "__pylav_original_cog_unload", None):
        return await discord.utils.maybe_coroutine(meth)


async def cog_before_invoke(self: DISCORD_COG_TYPE, context: PyLavContext):
    try:
        await self.pylav.wait_until_ready(timeout=30)
    except asyncio.TimeoutError as e:
        LOGGER.debug("Discarded command due to PyLav not being ready within 30 seconds")

        LOGGER.verbose(
            "Discarded command due to PyLav not being ready within 30 seconds - Guild: %s - Command: %s",
            context.guild,
            context.command.qualified_name,
        )

        raise CheckFailure(_("PyLav is starting up; please try again in a few minutes.")) from e
    if meth := getattr(self, "__pylav_original_cog_before_invoke", None):
        return await discord.utils.maybe_coroutine(meth)


async def initialize(self: DISCORD_COG_TYPE, *args, **kwargs) -> None:
    if not self.init_called:
        await self.pylav.register(self)
        await self.pylav.initialize()
        self.init_called = True
    if meth := getattr(self, "__pylav_original_initialize", None):
        return await discord.utils.maybe_coroutine(meth, *args, **kwargs)


async def cog_check(self: DISCORD_COG_TYPE, context: PyLavContext) -> bool:
    # This cog mock discord objects and sends them on the listener
    #   Due to the potential risk for unexpected behaviour - disabled all commands if this cog is loaded.
    if any(context.bot.get_cog(name) is not None for name in INCOMPATIBLE_COGS):
        return False
    if not (getattr(context.bot, "pylav", None)):
        return False
    meth = getattr(self, "__pylav_original_cog_check", None)
    if not context.guild:
        return await discord.utils.maybe_coroutine(meth, context) if meth else True
    if getattr(context, "player", None):
        config = context.player.config
    else:
        config = context.bot.pylav.player_config_manager.get_config(context.guild.id)

    if (channel_id := await config.fetch_text_channel_id()) != 0 and channel_id != context.channel.id:
        return False
    return await discord.utils.maybe_coroutine(meth, context) if meth else True


def class_factory(
    bot: DISCORD_BOT_TYPE,
    cls: type[DISCORD_COG_TYPE],
    cogargs: tuple[object],
    cogkwargs: dict[str, object],
) -> DISCORD_COG_TYPE:  # sourcery no-metrics
    """
    Creates a new class which inherits from the given class and overrides the following methods:
    - cog_check
    - cog_unload
    - cog_before_invoke
    - initialize
    - cog_command_error
    """
    if not bot.get_command(pylav_credits.qualified_name):
        bot.add_command(pylav_credits)
    if not bot.get_command(pylav_version.qualified_name):
        bot.add_command(pylav_version)
    if not bot.get_command(pylav_sync_slash.qualified_name):
        bot.add_command(pylav_sync_slash)
    argspec = inspect.getfullargspec(cls.__init__)
    if ("bot" in argspec.args or "bot" in argspec.kwonlyargs) and bot not in cogargs:
        cogkwargs["bot"] = bot

    cog_instance = cls(*cogargs, **cogkwargs)
    if not hasattr(cog_instance, "__version__"):
        cog_instance.__version__ = "0.0.0"
    cog_instance.pylav = Client(bot=bot, cog=cog_instance, config_folder=cog_data_path(raw_name="PyLav"))
    cog_instance.bot = bot
    cog_instance.init_called = False
    cog_instance._init_task = cls.cog_check
    cog_instance.lavalink = cog_instance.pylav
    old_cog_on_command_error = cog_instance._get_overridden_method(cog_instance.cog_command_error)
    old_cog_unload = cog_instance._get_overridden_method(cog_instance.cog_unload)
    old_cog_before_invoke = cog_instance._get_overridden_method(cog_instance.cog_before_invoke)
    old_cog_check = cog_instance._get_overridden_method(cog_instance.cog_check)
    old_cog_initialize = getattr(cog_instance, "initialize", None)
    if old_cog_on_command_error:
        cog_instance.__pylav_original_cog_command_error = old_cog_on_command_error
    if old_cog_unload:
        cog_instance.__pylav_original_cog_unload = old_cog_unload
    if old_cog_before_invoke:
        cog_instance.__pylav_original_cog_before_invoke = old_cog_before_invoke
    if old_cog_check:
        cog_instance.__pylav_original_cog_check = old_cog_check
    if old_cog_initialize:
        cog_instance.__pylav_original_initialize = old_cog_initialize

    cog_instance.cog_command_error = MethodType(cog_command_error, cog_instance)
    cog_instance.cog_unload = MethodType(cog_unload, cog_instance)
    cog_instance.cog_before_invoke = MethodType(cog_before_invoke, cog_instance)
    cog_instance.initialize = MethodType(initialize, cog_instance)
    cog_instance.cog_check = MethodType(cog_check, cog_instance)

    return cog_instance


async def pylav_auto_setup(
    bot: DISCORD_BOT_TYPE,
    cog_cls: type[DISCORD_COG_TYPE],
    cogargs: tuple[object, ...] = None,
    cogkwargs: dict[str, object] = None,
    initargs: tuple[object, ...] = None,
    initkwargs: dict[str, object] = None,
) -> DISCORD_COG_TYPE:
    """Injects all the methods and attributes to respect PyLav Settings and keep the user experience consistent.

    Adds `.bot` attribute to the cog instance.
    Adds `.pylav` attribute to the cog instance and starts up PyLav
    Overwrites cog_unload method to unregister the cog from Lavalink,
        calling the original cog_unload method once the PyLav unregister code is run.
    Overwrites cog_before_invoke
        To force commands to wait for PyLav to be ready
    Overwrites cog_check method to check if the cog is allowed to run in the current context,
        If called within a Guild then we check if we can run as per the PyLav Command channel lock,
        if this check passes then the original cog_check method is called.
    Overwrites cog_command_error method to handle PyLav errors raised by the cog,
        if the cog defines their own cog_command_error method,
        this will still be called after the built-in PyLav error handling if the error raised was unhandled.
    Overwrites initialize method to handle PyLav startup,
        calling the original initialize method once the PyLav initialization code is run, if such method exists. code
        is run.


    Args:
        bot (DISCORD_BOT_TYPE): The bot instance to load the cog instance to.
        cog_cls (type[DISCORD_COG_TYPE]): The cog class load.
        cogargs (tuple[object]): The arguments to pass to the cog class.
        cogkwargs (dict[str, object]): The keyword arguments to pass to the cog class.
        initargs (tuple[object]): The arguments to pass to the initialize method.
        initkwargs (dict[str, object]): The keyword arguments to pass to the initialize method.

    Returns:
        DISCORD_COG_TYPE: The cog instance loaded to the bot.

    Example:
        >>> from pylav.extension.red.utils.required_methods import pylav_auto_setup
        >>> from discord.ext.commands import Cog
        >>> class MyCogClass(Cog):
        ...     def __init__(self, bot: DISCORD_BOT_TYPE, special_arg: object):
        ...         self.bot = bot
        ...         self.special_arg = special_arg


        >>> async def setup(bot: DISCORD_BOT_TYPE) -> None:
        ...     await pylav_auto_setup(bot, MyCogClass, cogargs=(), cogkwargs=dict(special_arg=42), initargs=(),
        initkwargs=dict())

    """
    if any(bot.get_cog(name) is not None for name in INCOMPATIBLE_COGS if (_name := name)):
        raise IncompatibleException(
            f"{_name} is loaded, this cog is incompatible with PyLav - PyLav will not work as long as this cog is loaded"
        )
    if cogargs is None:
        cogargs = ()
    if cogkwargs is None:
        cogkwargs = {}
    if initargs is None:
        initargs = ()
    if initkwargs is None:
        initkwargs = {}
    with _LOCK:
        cog_instance = class_factory(bot, cog_cls, cogargs, cogkwargs)
        await bot.add_cog(cog_instance)
    cog_instance._init_task = asyncio.create_task(cog_instance.initialize(*initargs, **initkwargs))
    cog_instance._init_task.add_done_callback(_done_callback)
    return cog_instance



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class LowPass(FilterMixin):
    __slots__ = ("_smoothing", "_default")

    def __init__(self, smoothing: float | None = None) -> None:
        super().__init__()
        self.smoothing = smoothing

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "smoothing": self.smoothing,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> LowPass:
        return cls(smoothing=data["smoothing"])

    def __repr__(self) -> str:
        return f"<LowPass: smoothing={self.smoothing}>"

    @property
    def smoothing(self) -> float | None:
        return self._smoothing

    @smoothing.setter
    def smoothing(self, v: float | None) -> None:
        self._smoothing = v

    @classmethod
    def default(cls) -> LowPass:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.smoothing is not None:
            response["smoothing"] = self.smoothing
        return response

    def reset(self) -> None:
        self.smoothing = None



from __future__ import annotations

from pathlib import Path
from typing import Literal

import discord
from redbot.core.i18n import Translator

from pylav.extension.red.ui.selectors.options.queue import QueueTrackOption, SearchTrackOption
from pylav.helpers.format.strings import shorten_string
from pylav.players.tracks.obj import Track
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class QueueSelectTrack(discord.ui.Select):
    def __init__(
        self,
        options: list[QueueTrackOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        interaction_type: Literal["remove", "play"],
        mapping: dict[str, Track],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.interaction_type = interaction_type
        self.mapping = mapping

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        track_id = self.values[0]
        track: Track = self.mapping.get(track_id)
        if track is None:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(description="Track not found", messageable=interaction),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return
        player = self.cog.pylav.get_player(interaction.guild)
        if not player:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    description="Player has been disconnected", messageable=interaction
                ),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return

        await interaction.response.defer(ephemeral=True)

        index = player.queue.index(track)
        index += 1
        if self.interaction_type == "remove":
            await self.cog.command_remove.callback(
                self.cog, interaction, track_url_or_index=f"{index}", remove_duplicates=True
            )
        else:
            await self.cog.command_bump.callback(self.cog, interaction, queue_number=index)
        self.view.stop()
        await self.view.on_timeout()


class SearchSelectTrack(discord.ui.Select):
    def __init__(
        self,
        options: list[SearchTrackOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        mapping: dict[str, Track],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.mapping = mapping

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        track_id = self.values[0]
        track: Track = self.mapping.get(track_id)

        if track is None:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(messageable=interaction, title=_("Track was not found.")),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return

        await self.cog.command_play.callback(
            self.cog,
            interaction,
            query=[track.uri],
        )
        self.view.stop()
        await self.view.on_timeout()



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Distortion(FilterMixin):
    __slots__ = (
        "_sin_offset",
        "_sin_scale",
        "_cos_offset",
        "_cos_scale",
        "_tan_offset",
        "_tan_scale",
        "_offset",
        "_scale",
        "_default",
    )

    def __init__(
        self,
        sin_offset: float | None = None,
        sin_scale: float | None = None,
        cos_offset: float | None = None,
        cos_scale: float | None = None,
        tan_offset: float | None = None,
        tan_scale: float | None = None,
        offset: float | None = None,
        scale: float | None = None,
    ) -> None:
        super().__init__()
        self.sin_offset = sin_offset
        self.sin_scale = sin_scale
        self.cos_offset = cos_offset
        self.cos_scale = cos_scale
        self.tan_offset = tan_offset
        self.tan_scale = tan_scale
        self.offset = offset
        self.scale = scale

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "sin_offset": self.sin_offset,
            "sin_scale": self.sin_scale,
            "cos_offset": self.cos_offset,
            "cos_scale": self.cos_scale,
            "tan_offset": self.tan_offset,
            "tan_scale": self.tan_scale,
            "offset": self.offset,
            "scale": self.scale,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> Distortion:
        return cls(
            sin_offset=data["sin_offset"],
            sin_scale=data["sin_scale"],
            cos_offset=data["cos_offset"],
            cos_scale=data["cos_scale"],
            tan_offset=data["tan_offset"],
            tan_scale=data["tan_scale"],
            offset=data["offset"],
            scale=data["scale"],
        )

    def __repr__(self):
        return (
            "<Distortion: "
            f"sin_offset={self.sin_offset}, "
            f"sin_scale={self.sin_scale}, "
            f"cos_offset={self.cos_offset}, "
            f"cos_scale={self.cos_scale}, "
            f"tan_offset={self.tan_offset}, "
            f"tan_scale={self.tan_scale}, "
            f"offset={self.offset}, "
            f"scale={self.scale}>"
        )

    @property
    def sin_offset(self) -> float | None:
        return self._sin_offset

    @sin_offset.setter
    def sin_offset(self, v: float | None) -> None:
        self._sin_offset = v

    @property
    def sin_scale(self) -> float | None:
        return self._sin_scale

    @sin_scale.setter
    def sin_scale(self, v: float | None) -> None:
        self._sin_scale = v

    @property
    def cos_offset(self) -> float | None:
        return self._cos_offset

    @cos_offset.setter
    def cos_offset(self, v: float | None) -> None:
        self._cos_offset = v

    @property
    def cos_scale(self) -> float | None:
        return self._cos_scale

    @cos_scale.setter
    def cos_scale(self, v: float | None) -> None:
        self._cos_scale = v

    @property
    def tan_offset(self) -> float | None:
        return self._tan_offset

    @tan_offset.setter
    def tan_offset(self, v: float | None) -> None:
        self._tan_offset = v

    @property
    def tan_scale(self) -> float | None:
        return self._tan_scale

    @tan_scale.setter
    def tan_scale(self, v: float | None) -> None:
        self._tan_scale = v

    @property
    def offset(self) -> float | None:
        return self._offset

    @offset.setter
    def offset(self, v: float | None) -> None:
        self._offset = v

    @property
    def scale(self) -> float | None:
        return self._scale

    @scale.setter
    def scale(self, v: float | None) -> None:
        self._scale = v

    @classmethod
    def default(cls) -> Distortion:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.sin_offset is not None:
            response["sinOffset"] = self.sin_offset
        if self.sin_scale is not None:
            response["sinScale"] = self.sin_scale
        if self.cos_offset is not None:
            response["cosOffset"] = self.cos_offset
        if self.cos_scale is not None:
            response["cosScale"] = self.cos_scale
        if self.tan_offset is not None:
            response["tanOffset"] = self.tan_offset
        if self.tan_scale is not None:
            response["tanScale"] = self.tan_scale
        if self.offset is not None:
            response["offset"] = self.offset
        if self.scale is not None:
            response["scale"] = self.scale
        return response

    def reset(self) -> None:
        self.sin_scale = (
            self.cos_offset
        ) = self.cos_scale = self.tan_offset = self.tan_scale = self.offset = self.scale = self.sin_offset = None



from __future__ import annotations

import dataclasses
from datetime import datetime


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkError:
    timestamp: int | datetime
    status: int
    error: str
    message: str
    path: str
    trace: str | None = None

    def __post_init__(self) -> None:
        if isinstance(self.timestamp, int):
            object.__setattr__(self, "timestamp", datetime.fromtimestamp(self.timestamp / 1000))

    def __bool__(self) -> bool:
        return False



from __future__ import annotations

from typing import TYPE_CHECKING

from dacite import from_dict
from yarl import URL

from pylav.compat import json
from pylav.extension.flowery.lyrics.responses import Error, Lyrics, TrackList

if TYPE_CHECKING:
    from pylav.core.client import Client
    from pylav.extension.flowery.base import FloweryAPI


class LyricsAPI:
    def __init__(self, client: Client, wrapper: FloweryAPI) -> None:
        self._wrapper = wrapper
        self._base_url = URL("https://api.flowery.pw/v1/lyrics")
        self._session = wrapper._cached_session
        self.client = client

    async def get_lyrics(self, query: str = None, isrc: str = None, spotify_id: str = None) -> Lyrics | Error:
        params = {}
        if query:
            params["query"] = query
        if isrc:
            params["isrc"] = isrc
        if spotify_id:
            params["spotify_id"] = spotify_id
        if not params:
            raise ValueError("You must provide either a query, isrc or spotify_id")

        async with self._session.get(
            self._base_url,
            params=params,
        ) as response:
            match response.status:
                case 200 | 202:
                    return from_dict(data_class=Lyrics, data=await response.json(loads=json.loads))
                case 404 | 422 | 500:
                    return from_dict(data_class=Error, data=await response.json(loads=json.loads))
                case __:
                    raise ValueError(f"Unexpected status code: {response.status}")

    async def search_lyrics(self, query: str) -> TrackList | Error:
        async with self._session.get(
            self._base_url / "search",
            params={"query": query},
        ) as response:
            match response.status:
                case 200:
                    return from_dict(data_class=TrackList, data=await response.json(loads=json.loads))
                case 404 | 422 | 500:
                    return from_dict(data_class=Error, data=await response.json(loads=json.loads))
                case __:
                    raise ValueError(f"Unexpected status code: {response.status}")
